-- $LastChangedDate: 2008-09-02 20:29:21 -0400 (Tue, 02 Sep 2008) $
-- Actor classes.
-- See module.txt for documentation and pitfalls.

------------------------------
-- Player class (singleton) --
------------------------------

Player = {}

-- Make (or remake) the player.
-- If aircraftType is nil, determine from cmd-line args.
function Player.Make( aircraftType )
	local self = Player

	-- Remake?
	if ( self.aircraft ) then
		sim.World:DetachObject( self.aircraft )
		self.aircraft = nil
	end

	if ( aircraftType ) then
		self.aircraftType = aircraftType
	else
		-- Parse cmdline args to select aircraft for player.
		self.aircraftType = sim.Aircraft.type_F18  -- default
		for k,v in pairs(sim.argv2) do
			if ( sim.Aircraft.typeArgMap[k] ) then
				self.aircraftType = sim.Aircraft.typeArgMap[k]
			end
		end
	end

	-- Create the player's aircraft.
	-- Correct position will be assigned later.
	self.aircraft = sim.Aircraft.new( Vec3(0,0,0),
									  self.aircraftType,
									  sim.ModelCache.LOAD_COPY )
	sim.World:AttachObject( self.aircraft )
	self.aircraft:SetState( sim.Craft.eState_PARKED ) -- default
	sim.Control.SetCurrentCraft( self.aircraft )

	-- Enable coordinated-turning if this is a propeller aircraft.
	if ( sim.argv2["-coturn"] ) then
		self.aircraft:EnableCoordinatedTurn( true )
		Log( "Enabled coturn." )
	elseif ( sim.argv2["-no-coturn"] ) then
		self.aircraft:EnableCoordinatedTurn( false )
		Log( "Disabled coturn." )
	else
		self.aircraft:EnableCoordinatedTurn( self.aircraft:IfHasPropeller() )
	end
end
Player.Make()  -- do it

-- Save player state.
function Player.Save()
	local self = Player
	local craftPhysics = self.aircraft:GetCraftPhysics()
	self.playerSaved = {
		aircraftState = self.aircraft:GetState(),
		matrix        = self.aircraft:GetMatrix(),
		throttle      = self.aircraft:GetThrottle(),
		speed         = craftPhysics:ComputeSpeed(),
		landingGear   = self.aircraft:IfLandingGear(),
	}
end

-- Reset player.
function Player.Reset()
	local self = Player
		
	-- Restore player from the playerSaved table.
	local craftPhysics = self.aircraft:GetCraftPhysics()
	BeginPlacement( self.aircraft )  -- disable collisions
	self.aircraft:Reset()
	self.aircraft:SetState(       self.playerSaved.aircraftState )
	self.aircraft:SetMatrix(      self.playerSaved.matrix )
	self.aircraft:SetThrottle(    self.playerSaved.throttle )
	self.aircraft:SetLandingGear( self.playerSaved.landingGear )
	craftPhysics:SetSpeed(        self.playerSaved.speed )
	EndPlacement( self.aircraft )

	-- Reset views.
	ChaseView.Init( sim.argv2["-ossim"] )
	sim.SuperView.Reset()

	-- Reset lights.
	Brightness.Reset()
end
Events.RegisterReset( Player.Reset )

------------------
-- Friend class --
------------------

Friend = {}

function Friend:new( aircraftType )
	local t = {
		aircraftType=aircraftType,
		aircraft = sim.Aircraft.new( Vec3(0,0,0), aircraftType, sim.ModelCache.LOAD_COPY )
	}
	setmetatable( t, self )
	self.__index = self
	return t
end
