-- $LastChangedDate: 2008-09-02 20:29:21 -0400 (Tue, 02 Sep 2008) $
-- Game code.
-- See module.txt for documentation and pitfalls.

-- Randomize.
math.randomseed( os.time() )

----------------
-- Game class --
----------------
Game = {}

-- There are different kinds of scores.
Game.SCORE_NONE    = false
Game.SCORE_CRASH   = 1
Game.SCORE_LANDING = 2
Game.score         = Game.SCORE_NONE
Game.playerReplay  = Replay:new( Player.aircraft )

-- Reset.
function Game.Reset()
	local self = Game
	self.score = Game.SCORE_NONE

	-- SUBTLE: Although the Replay object is reassigned and no longer referenced,
	-- Lua delays garbage-collecting it.  Replay::Stop() is called to stop
	-- its timer-ticks so that it becomes inactive (zombie) else it would
	-- content with the new Replay object!
	self.playerReplay:Stop()  -- disable its timer-ticks (turn into zombie)
	self.playerReplay = Replay:new( Player.aircraft )  -- re-create
end
Events.RegisterReset( Game.Reset )

-- (private) Increment a type of score in binary configuration.
function Game.WriteScoreInt( key, inc )
	local self = Game

	-- Read score from ConfBin.
	local readOk,score = sim.ConfBin:ReadInt( key )
	if ( not readOk ) then
		-- This probably means the integer value isn't yet in the conf file.
		score = 0
	end

	-- Increment score.
	score = score + inc

	-- Write updated score to ConfBin.
	local writeOk = sim.ConfBin.WriteInt( key, score )
	if ( not writeOk ) then
		print( "ERROR [Lua] accessing score in bin conf file" )
	end
end

-- Pass score as Game.SCORE_* enum.
-- Returns true if state transition.
function Game.UpdateScore( score )
	local self = Game

	-- Did state change?
	if ( self.score ~= score ) then
		-- Remember state.
		self.score = score

		-- Update binary configuration file (~/.palomino.cfg).
		if ( score == self.SCORE_CRASH ) then
			self.WriteScoreInt( "game_score_crashes", 1 )
		elseif ( score == self.SCORE_LANDING ) then
			self.WriteScoreInt( "game_score_landings", 1 )
		end

		-- Return true to indicate the score changed.
		return true
	else
		-- Return false to indicate no change.
		return false
	end
end

-- Get Replay object.
function Game.GetPlayerReplay()
	local self = Game
	return self.playerReplay
end

----------------------------
-- RescueHelicopter class --
----------------------------
RescueHelicopter = {}

function RescueHelicopter.Start()
	local self = RescueHelicopter

	-- Already started?
	if ( self.rescueHelicopter ) then
		return
	end	

	-- Place rescue helicopter above player's burning wreck.
	local pos = Player.aircraft:GetPosition()
	pos[LAT] = pos[LAT] - sim.World.Meters2Sim(20.0)
	pos[LON] = pos[LON] - sim.World.Meters2Sim(20.0)
	pos[ALT] = pos[ALT] + sim.World.Meters2Sim(20.0)
	self.rescueHelicopter = sim.Aircraft.new( pos,
		 									  sim.Aircraft.type_Sikorsky,
											  sim.ModelCache.LOAD_SHARED )

	-- Play rotor sound.
	sim.Sound.Play( "rotor.wav.gz", sim.Sound.PLAY_LOOP, 0.6 )

	-- CraftPhysics is limited to airplanes
	self.rescueHelicopter:GetCraftPhysics():Enable( false )
	self.rescueHelicopter:SetThrottle( 1.0 )
	self.rescueHelicopter:SetCollidable( false )
	self.rescueHelicopter:Rotate( AXIS_YAW, Deg2Rad(270.0) );
	sim.World.AttachObject( self.rescueHelicopter )

	-- Change view mode.
	sim.SuperView.SetViewMode( sim.SuperView.eViewMode_CHASE )
	ChaseView.chaseViewpoint = ChaseView.chaseViewpointPlayer
	ChaseView.Reset()
	local distance = sim.World.Meters2Sim( 3.5 * ChaseView.distance )
	ChaseView.chaseViewpoint:SetDistance( 0.0, 0.0, distance )
	ChaseView.chaseViewpoint:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_VERT,
									 Deg2Rad(-25.0) )
end

function RescueHelicopter.Stop()
	local self = RescueHelicopter
	if ( self.rescueHelicopter ) then
		sim.World.DetachObject( self.rescueHelicopter )
		self.rescueHelicopter = nil
	end
end

-- For rescue helicopter to disappear after program reset.
Events.RegisterReset( RescueHelicopter.Stop )
