-- $LastChangedDate: 2008-09-04 10:49:45 -0400 (Thu, 04 Sep 2008) $
-- Mission (aircraft carrier at Golden Gate).
-- mission.lua defines an equivalent of a Mission base class which this extends.
-- See module.txt for documentation and pitfalls.

-- Specialize mission class.
assert( Mission )  -- mission.lua defines base methods
Mission.BaseReset  = Mission.Reset
Mission.friends    = {}
Mission.viewCnt    = 1
Mission.playerPos  = Vec3( 8110.0, -16500.0, -500.0 )
Mission.carrierPos = Vec3( 8000.0, -15500.0, -550.0 )
Mission.light = {
	pos      = Vec3( -15000.0, -500.0, 70000.0 ),
	ambient  = RGBA(0.20, 0.20, 0.20, 1.00),
	diffuse  = RGBA(0.60, 0.60, 0.60, 1.00),
	specular = RGBA(0.30, 0.30, 0.30, 1.00),
}

-- Place aircraft on carrier, tilted up appropriately.
function Mission.PlaceAircraftOnCarrier( aircraft, carrier, carrierSlot )
	local self = Mission

	carrier:PlaceAircraft( aircraft, sim.Carrier.slots[carrierSlot] )
	if ( aircraft:IfHasPropeller() ) then
		aircraft:Rotate( AXIS_PITCH, Deg2Rad(10.0) )
	else
		aircraft:Rotate( AXIS_PITCH, Deg2Rad(0.5) )
	end
end

-- Decorate carrier with aircraft.
function Mission.DecorateCarrier( carrier, aircraftType )
	local self = Mission

	local firstCarrier = (carrier == carriers[1])
	for k,v in pairs(sim.Carrier.slots) do
		-- Leave the first few slots empty (reserved for player).
		if ( not firstCarrier or k > sim.Carrier.takeoffSlotCnt ) then
			local aircraft = sim.Aircraft.new( Vec3(0,0,0), aircraftType, sim.ModelCache.LOAD_SHARED )
			local craftPhysics = aircraft:GetCraftPhysics()
			craftPhysics:Enable( false )  -- don't need to compute physics
			sim.World:AttachObject( aircraft )
			aircraft:Rotate( AXIS_YAW, RADIAN_90 )
			-- NOTE: Finish any orientation before PlaceAircraftOnCarrier().
			self.PlaceAircraftOnCarrier( aircraft, carrier, k )
		end
	end
end

-- Stop animation of friendly aircraft.
function Mission.StopFriendReplay()
	local self = Mission

	-- Call Replay::Stop() since Lua's garbage-collector will delay destroying!
	if ( self.friendReplay ) then
		self.friendReplay:Stop()
	end
	self.friendReplay = nil
end

-- Create friend.
function Mission.MakeFriends()
	local self = Mission

	self.friends = {}  -- recreate

	local friend = Friend:new( sim.Aircraft.type_F18 )
	table.insert( self.friends, friend )
	self.carrier0:PlaceAircraft( friend.aircraft, sim.Carrier.slots[self.friendSlot] )
	sim.World:AttachObject( friend.aircraft )
	sim.World.RotateTangential( friend.aircraft, friend.aircraft:GetPosition() )
	friend.aircraft:Rotate( AXIS_YAW, Deg2Rad(8.5) )
	friend.aircraft:SetLandingGear( true )

	-- Keep orientation of auxillary chase-plane after subsequent program resets.
	if ( ChaseView.chaseViewpointFriend ) then
		ChaseView.chaseViewpointFriend:SetTarget( friend.aircraft )
		if ( self.ifDoneFriendViewpoint == nil ) then
			self.ifDoneFriendViewpoint = true
			ChaseView.chaseViewpointFriend:SetDistance( 0.0, 0.0, sim.World.Meters2Sim(ChaseView.distance) )
			ChaseView.chaseViewpointFriend:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_HORZ, ChaseView.horzAngle )
			ChaseView.chaseViewpointFriend:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_VERT, ChaseView.vertAngle )
		end
		--  Save state for when chase-plane is reset.
		ChaseView.chaseViewpointFriend:Save()
	end

	-- Animate friend using replay data.
	self.StopFriendReplay()
	self.friendReplay = Replay:new( friend.aircraft )
	self.friendReplay:Load( self.friendReplayFilename )
	self.friendReplay:Play()
end

-- They served their purpose.
function Mission.KillFriends()
	local self = Mission

	if ( self.friends and self.friends[1] ) then
		self.StopFriendReplay()

		sim.World.DetachObject( self.friends[1].aircraft )
		self.friends[1] = nil
	end
end

-- Reset mission.
function Mission.Reset()
	local self = Mission

	-- Call base Reset().
	self.BaseReset()

	-- RESTART command is generated by the user changing mission by GUI.
	local cmdFields = CommandQueue.GetCmdFields()
	if ( cmdFields and (cmdFields[CommandQueue.COMMAND_FIELD] == "RESTART") ) then
		local missionName = cmdFields[CommandQueue.RESTART_FIELD_MISSION]
		if ( missionName == "takeoff" ) then
			self.mode = self.MODE_TAKEOFF
		elseif ( missionName == "land" ) then
			self.mode = self.MODE_LAND
		end
	end

	-- Reset HUD.
	if ( self.mode == self.MODE_LAND ) then
		hudStringInfo:SetText( "Land on carrier." )
	else
		hudStringInfo:SetText( "Takeoff." )
	end

	-- Reset player.
	Player.aircraft:Reset()
	local craftPhysics = Player.aircraft:GetCraftPhysics()
	if ( self.playerSlot == 3 ) then
		-- Front slot.
		Player.aircraft:Rotate( AXIS_YAW, Deg2Rad(2.5) )  -- yaw left
	else
		-- Rear slots.
		Player.aircraft:Rotate( AXIS_YAW, Deg2Rad(8.5) )  -- yaw left
	end
	if ( self.mode == self.MODE_LAND ) then
		Player.aircraft:SetState( sim.Craft.eState_FLYING )
		Player.aircraft:SetThrottle( 0.3 )
		craftPhysics:SetSpeed( sim.Physics.KPH2Speed( 125.0 ) )

		BeginPlacement( Player.aircraft )  -- disable collision
		Player.aircraft:SetPosition( self.playerPos )
		EndPlacement( Player.aircraft )  -- enable collision
	else
		Player.aircraft:SetState( sim.Craft.eState_PARKED )
		Player.aircraft:SetThrottle( 0.0 )
		craftPhysics:SetSpeed( sim.Physics.KPH2Speed( 0.0 ) )

		self.PlaceAircraftOnCarrier( Player.aircraft, self.carrier0, self.playerSlot )
	end

	-- Reset view.
	sim.SuperView.SetViewMode( sim.SuperView.eViewMode_OTW_FRONT )
	if ( self.mode ~= self.MODE_TAKEOFF ) then
		-- Adjust chase-plane so that hudstring is readable.
		ChaseView.chaseViewpoint:Rotate( sim.ChaseViewpoint.AXIS_CHASE_PLANE_VERT, Deg2Rad(5.0) )
	end
    if ( ChaseView.chaseViewpointFriend ) then  -- unless -fast
		ChaseView.chaseViewpointFriend:Reset()
    end

	-- Make/reset friendly aircraft.
	self.KillFriends()
	self.MakeFriends()
end

-- Initialization/start mission (done once).
function Mission.Start()
	local self = Mission

	-- Specialize mission class.
	-- Initial position.
	self.initialPos = sim.World:SphereVertex2WorldVertex( 37.518800077449, -122.37326847224, 2.0 )
	if ( sim.argv2["-front-slot"] ) then
		self.playerSlot = 3
		self.friendSlot = 2  -- rear slot
		self.friendReplayFilename = "replays/replayTakeoffRearSlot.dat.gz"
	else
		self.playerSlot = 1
		self.friendSlot = 3  -- front slot
		self.friendReplayFilename = "replays/replayTakeoffFrontSlot.dat.gz"
	end

	-- Initialize sky and lights.
	sim.Sky:SetLightColor( self.light.ambient, self.light.diffuse, self.light.specular )
	sim.Sky:SetLightPosition( self.light.pos )

	-- How to begin the mission.
	self.MODE_TAKEOFF = 1
	self.MODE_LAND    = 2
	self.mode         = self.MODE_TAKEOFF  -- default
	if ( sim.argv2["-takeoff"] ) then
		self.mode = self.MODE_TAKEOFF
	elseif ( sim.argv2["-land"] ) then
		self.mode = self.MODE_LAND
	end

	-- Initialize views.
	ChaseView.chaseViewpoint = ChaseView.chaseViewpointPlayer
	if ( not sim.argv2["-fast"] ) then
		self.viewCnt = 2  -- disable view of friend
	end
	if ( self.viewCnt == 3 ) then
		-- Create a view to show a friendly aircraft.
		local sx,sy = sim.Window.GetWindowSize()
		ChaseView.chaseViewpointFriend = sim.ChaseViewpoint.new()
		self.view3 = sim.View.new( sx * 0.75, sy * 0.75, sx, sy )
		self.view3:SetViewpoint( ChaseView.chaseViewpointFriend )
		self.view3:SetOutlineColor( defs.viewOutlineColorDisabled )
	end
	ChaseView.chaseViewpoint:Save()
	MakeSecondView()

	-- Create aircraft carrier(s).
	self.carriers = {
		sim.Carrier.new( self.carrierPos, 0 ),
	}
	self.carrier0 = self.carriers[1]

	-- Info.
	hudStringInfo = sim.HudString.new( { text="", orientation="TOP_LEFT" } )
	hudStringInfo:SetColor( 1.0, 1.0, 0.6, 1.0 )
	hudStringFriend = sim.HudString.new( { text="", orientation="BOTTOM_RIGHT" } )
	hudStringFriend:SetFontSize( 24.0 )

	-- Initial reset.
	self.Reset()

	-- Load scene.
	assert( Scene.Load( "Golden Gate" ) )
end

-- Mission-specific timer-tick handler.
function Mission.Tick()
	local self = Mission

	-- Did the player land on the carrier?
	if ( self.mode == self.MODE_LAND ) then
		local state = Player.aircraft:GetState()
		if ( state == sim.Craft.eState_LANDING ) then
			local playerSpeedKPH = AircraftComputeSpeedKPH( Player.aircraft )
			if ( playerSpeedKPH > 3.0 ) then
				hudStringInfo:SetText( "Landing..." )
			else
				-- This updates only on state transition.
				Game.UpdateScore( Game.SCORE_LANDING )
				hudStringInfo:SetText( "Success!" )
				-- Switch state.
				Player.aircraft:SetState( sim.Craft.eState_PARKED )
			end
		elseif ( state == sim.Craft.eState_PARKED ) then
			hudStringInfo:SetText( "Success!" )
		elseif ( state == sim.Craft.eState_FLYING ) then
			hudStringInfo:SetText( "Land on carrier." )
		end
	end

	-- Animate UFO.
	-- UFO impossibly changes direction ;-)
	-- The meter constants are from Scene.
	if ( Scene.objects.ufo ) then
		local pos = Scene.objects.ufo.obj:GetPosition()
		if ( self.mode == self.MODE_TAKEOFF ) then
			if ( pos[LON] < self.initialPos[LON] + M2W(250.0) ) then
				pos[LON] = pos[LON] + M2W(2.5)
			else
				if ( pos[ALT] < self.initialPos[ALT] + M2W(100.0) ) then
					pos[ALT] = pos[ALT] + M2W(1.5)
				else
					pos[LAT] = pos[LAT] - M2W(2.5)
				end
			end
		else
			if ( pos[LON] < self.initialPos[LON] + M2W(-200.0) ) then
				pos[LON] = pos[LON] + M2W(4.0)
			else
				pos[LAT] = pos[LAT] + M2W(4.0)
			end
		end
		Scene.objects.ufo.obj:SetPosition( pos )
	end
	
end

-- Let startup.lua continue...
