/*****************************************************************************
 * $LastChangedDate: 2008-09-25 16:01:29 -0400 (Thu, 25 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	HUD.
 * @remarks
 * In OSG, multiple lines of text is drawn by attaching nodes of
 * osg::Text objects, contained in Geodes, that define different
 * 2D positions and strings.  These are abstracted by the HudString class.
 *
 * Arcane magic and kludges:
 * -------------------------
 * - Constant window size:
 *   Indeed, the width,height of the HUD remains constant at 1280,1024.
 *   Intentionally, there is no listener for a window-resize event,
 *   since OSG will automatically scale osgText strings.
 * - Right-alignment:
 *   This is done using HUD_FONT_WIDTH_RATIO which is a magic value
 *   obtained by experimentation.  It is specific to the "arial.ttf" font.
 *   Using another font while orienting at right would probably result
 *   in misalignment.  Though it isn't used for normal left-alignment.
 *
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define HUD_HUD_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/osg.hh"
using namespace glue;
#include "graph/module.hh"
using namespace graph;
#include "view/module.hh"
using namespace view;
#include "hud/module.hh"

#include <osg/CameraNode>

namespace hud {

INTERN const fp			HUD_COLOR_ALPHA = 1.0f;
INTERN const osg::Vec4f	HUD_COLOR( 0.8f, 1.0f, 0.8f, HUD_COLOR_ALPHA );
INTERN const string		HUD_FONT_NAME( "arial.ttf" );
INTERN const fp			HUD_FONT_SIZE = 28.0f;
INTERN const fp         HUD_FONT_WIDTH_RATIO  = 0.593f;  // [see issues] specific to "fonts/arial.ttf"
INTERN const fp         HUD_FONT_HEIGHT_RATIO = 1.0f;
INTERN const fp			HUD_ORTHO_X1 = 0;
INTERN const fp			HUD_ORTHO_X2 = 1280;  // indeed, HUD size remains constant
INTERN const fp			HUD_ORTHO_Y1 = 0;
INTERN const fp			HUD_ORTHO_Y2 = 1024;

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  HudString::Args  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor.
 *****************************************************************************/
Hud::HudString::Args::Args( void )
:	mView(GET_WINDOW().GetMainView()),
	mPos(0.0f,0.0f),
	mOrientation(eOrientation_BOTTOM_LEFT),
	mText(),
	mFontName(HUD_FONT_NAME),
	mFontSize(HUD_FONT_SIZE),
    mColor(HUD_COLOR)
{
	// NOP
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  HudString  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Hud::HudString::HudString( const Args& args )
:	mTextObj(new osgText::Text),
	mView(args.mView)
{
// Empty mText is ok.
ASSERT( args.mPos[XX] >= 0 and args.mPos[XX] < 50000 );
ASSERT( args.mPos[YY] >= 0 and args.mPos[YY] < 50000 );
ASSERT( args.mFontSize > 0 );
ASSERT( not args.mFontName.empty() );

	SET_TYPESIG(this,TYPESIG_HUDSTRING);

	// Set the orientation/alignment.
	const Vector2 pos = OrientPosition( args );
    mTextObj->setPosition( osg::Vec3( pos[XX], pos[YY], 0.0f ) );
	mTextObj->setText( args.mText );
    mTextObj->setFont( base::GetDataDir() + string("/fonts/") + args.mFontName );
    mTextObj->setCharacterSize( args.mFontSize, 1.0f );
	mTextObj->setColor( RGBA2VEC4F(args.mColor) );
}

Hud::HudString::~HudString()
{
	INVALIDATE_TYPESIG(this,TYPESIG_HUDSTRING);
}

/*****************************************************************************
 * (private ctor subroutine)
 * Re-orient the position of a HudString.
 * The input position must be oriented at BOTTOM_LEFT.
 *****************************************************************************/
Vector2
Hud::HudString::OrientPosition( const Args& args )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

	Vector2 pos = args.mPos;

	// Estimate the width,height of the font.
	// KLUDGE: HUD_FONT_WIDTH_RATIO is specific to to "arial.ttf".
	const fp fontWidth  = args.mFontSize * HUD_FONT_WIDTH_RATIO;
	const fp fontHeight = args.mFontSize * HUD_FONT_HEIGHT_RATIO;

	// Computing Y is straightforward.
	// Computing X for the right-alignment cases,
	// requires accounting for how many chars the text string has.
	switch ( args.mOrientation )
	{
		case eOrientation_BOTTOM_LEFT: default:
		{
			// NOP
		}
		break;

		case eOrientation_BOTTOM_RIGHT:
		{
			pos[XX] = HUD_ORTHO_X2 - pos[XX] - fontWidth * args.mText.size();
		}
		break;

		case eOrientation_TOP_LEFT:
		{
			pos[YY] = HUD_ORTHO_Y2 - pos[YY] - fontHeight;
		}
		break;

		case eOrientation_TOP_RIGHT:
		{
			pos[XX] = HUD_ORTHO_X2 - pos[XX] - fontWidth * args.mText.size();
			pos[YY] = HUD_ORTHO_Y2 - pos[YY] - fontHeight;
		}
		break;

		case eOrientation_CENTER:
		{
			pos[XX] = HUD_ORTHO_X2 * 0.5f - (fontWidth * args.mText.size() * 0.5f);
			pos[YY] = HUD_ORTHO_Y2 * 0.5f + pos[YY];  // client may add Y offset
		}
		break;
	}

	return pos;
}

/*****************************************************************************
 * Change the value of HudString.
 *****************************************************************************/
void
Hud::HudString::SetText( const char text[] )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

	mTextObj->setText( text );
}

void
Hud::HudString::SetText( const string& text )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

	mTextObj->setText( text.c_str() );
}

/*****************************************************************************
 * Set position by transforming 3D vertex into 2D.
 *****************************************************************************/
void
Hud::HudString::SetPosition3D( const WorldVertex& pos3 )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

	// Transform 3D --> 2D.
	Vector2 pos2 = mView->Project( pos3 );

	// Remap 2D from viewport's dimensions to HUD's dimensions.
	const Rect<int> viewport = mView->GetRect();
	pos2[XX] = Remap( pos2[XX], viewport.x2, HUD_ORTHO_X2 );
	pos2[YY] = Remap( pos2[YY], viewport.y2, HUD_ORTHO_Y2 );

	// Adjust so that text is slightly above aircraft.
	pos2[XX] -= 16.0f;
	pos2[YY] += 16.0f;

	// Directly change position of osg::Text.
    mTextObj->setPosition( osg::Vec3( pos2[XX], pos2[YY], 0.0f ) );
}

/*****************************************************************************
 * Set font size.
 *****************************************************************************/
void
Hud::HudString::SetFontSize( const fp fontSize )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

    mTextObj->setCharacterSize( fontSize, 1.0f );
}

/*****************************************************************************
 * Set color of text.
 *****************************************************************************/
void
Hud::HudString::SetColor( const RGBA rgba )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);

	mTextObj->setColor( RGBA2VEC4F(rgba) );
}

/*****************************************************************************
 * Set drawing text with 4 gradient colors.
 *****************************************************************************/
void
Hud::HudString::SetGradientColors4( const TinyArray<4,RGBA>& colors4 )
{
CHECK_TYPESIG(this,TYPESIG_HUDSTRING);
ASSERT( colors4.size() == 4 );

	mTextObj->setColorGradientCorners( RGBA2VEC4F(colors4[0]),
                                       RGBA2VEC4F(colors4[1]),
                                       RGBA2VEC4F(colors4[2]),
                                       RGBA2VEC4F(colors4[3]) );
	mTextObj->setColorGradientMode( osgText::Text::OVERALL );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////  Hud  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

Hud* Hud::msInstance;

/*****************************************************************************
 * ctor/dtor.
 * Root node is needed to place the Geodes containing osg::Text objects
 * (each HudString is a Node).
 *****************************************************************************/
Hud::Hud( void )
:	mCamera(new osg::CameraNode)
{
ASSERT_RUN_ONCE;

	// Attach to Window (where to render).
	GET_WINDOW().OsgAddCamera( mCamera.get(), Window::eCamera_2D );

	// Set 2D viewport.
	const WidthHeight<int> wh = GET_WINDOW().GetWindowSize();
    mCamera->setViewport( 0, 0, wh.w, wh.h );

    // Set a 2D projection matrix.
    mCamera->setProjectionMatrix( osg::Matrix::ortho2D( HUD_ORTHO_X1,
                                                        HUD_ORTHO_X2,
                                                        HUD_ORTHO_Y1,
                                                        HUD_ORTHO_Y2 ) );
    // Set the view matrix.
    mCamera->setReferenceFrame( osg::Transform::ABSOLUTE_RF );
    mCamera->setViewMatrix( osg::Matrix::identity() );

    // Only clear the depth buffer.
    mCamera->setClearMask( GL_DEPTH_BUFFER_BIT );

    // Draw subgraph after main camera view.
    mCamera->setRenderOrder( osg::CameraNode::POST_RENDER );
}

Hud::~Hud()
{
	// NOP
}

/*****************************************************************************
 * Enable/disable HUD.
 *****************************************************************************/
void
Hud::Enable( const bool enable )
{
	mCamera->setNodeMask( enable ? ~0 : 0 );
}

/*****************************************************************************
 * Attach a HudString.
 *****************************************************************************/
void
Hud::AttachHudString( shptr<HudString> hudString )
{
CHECK_TYPESIG(hudString,TYPESIG_HUDSTRING);

	// Create an Text object and a Geode to contain it.
    RefPtr<osg::Geode> geode = new osg::Geode;
    geode->addDrawable( hudString->mTextObj.get() );  // osgText::Text

	// Add geode to camera.
    mCamera->addChild( geode.get() );

	// Set stateset.
    geode->getOrCreateStateSet()->setMode( GL_LIGHTING, osg::StateAttribute::OFF );
}

/*****************************************************************************
 * @return HUD's node.
 *****************************************************************************/
RefPtr<osg::Node>
Hud::GetNode( void )
{
ASSERT( mCamera != NULL );

	// The osg::Text objects are children of the HUD's Camera.
	return mCamera.get();
}

} // namespace hud
