/*****************************************************************************
 * $LastChangedDate: 2008-08-31 13:03:54 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	HUD.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef HUD_HUD_HH
#define HUD_HUD_HH 1

#include "view/module.hh"
using namespace view;
#include <osgText/Text>
#include <osg/CameraNode>

namespace hud {

////////////////////////////////////////////////////////////////////////////////
/// @brief HUD.
///
/// Strings of text are displayed by the client allocating a HudString object.
/// HudString defines a 2D position and the value of the string.
/// HudString can be later modified or removed from the Hud.
///
/// HUD is auto-drawn by OSG since the HUD is comprised of a camera with
/// a 2D projection matrix and osg::Text objects which define a line of text.
///
class Hud
{
PREVENT_COPYING(Hud)
public:
	////////////////////////////////////////////////////////////////////////////
	/// @brief Defines a string of text to be displayed in HUD.
	///
	class HudString : public Shared
	{
	PREVENT_COPYING(HudString)
	friend class Hud;
	public:
		enum eOrientation
		{
			eOrientation_BOTTOM_LEFT,   // default
			eOrientation_BOTTOM_RIGHT,  // same order as OSG setColorGradientCorners()
			eOrientation_TOP_LEFT,
			eOrientation_TOP_RIGHT,
			eOrientation_CENTER,
		};

		// For setting RGBA array passed to SetGradientColors().
		enum eGradientColor
		{
			eGradientColor_TOP_LEFT,
            eGradientColor_BOTTOM_LEFT,
            eGradientColor_BOTTOM_RIGHT,
            eGradientColor_TOP_RIGHT,
		};

		/// @brief ctor args for HudString().
		///
		struct Args
		{
			Args( void );
			shptr<View>		mView;
			Vector2			mPos;
			eOrientation	mOrientation;
			string			mText;
			string			mFontName;
			fp				mFontSize;
			RGBA			mColor;
		};
					HudString( const Args& args );
					~HudString();
		void		SetText( const char text[] );
		void		SetText( const string& text );
		void		SetPosition3D( const WorldVertex& pos3 );
		void		SetFontSize( const fp fontSize );
		void		SetColor( const RGBA rgba );
		void		SetGradientColors4( const TinyArray<4,RGBA>& colors4 );

	private:
		Vector2		OrientPosition( const Args& args );

	private:
		RefPtr<osgText::Text>	mTextObj;	///< underlying OSG object
		shptr<View>				mView;		///< the View this HUD string is rendered inside
	public:
		DECL_TYPESIG(TYPESIG_HUDSTRING)
	};

private:
                        Hud( void );
                        ~Hud();

public:
	CLASS_METHOD Hud&	GetInstance( void )  // Singleton
	{
		if ( UX( msInstance == NULL ) )
			msInstance = new Hud;
		return *msInstance;
	}

	void			Enable( const bool enable = true );
	RefPtr<Node>	GetNode( void );
	void			AttachHudString( shptr<HudString> hudString );

private:
	CLASS_VAR Hud*			msInstance;
	RefPtr<osg::CameraNode>	mCamera;		///< contains the osg::Text objects
};

} // namespace hud

#endif // HUD_HUD_HH
