/*****************************************************************************
 * $LastChangedDate: 2008-08-29 01:40:45 -0400 (Fri, 29 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "program/module.hh"
#include "program/aircraft.hh"
using namespace program;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "Aircraft"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

#if ! DOXYGEN
// Need a function to check typesig because of macro/template limitations.
INTERN void
CheckTypesigAircraft( shptr<Aircraft> obj )
{
	CHECK_TYPESIG(obj,TYPESIG_AIRCRAFT);
}
#endif // ! DOXYGEN

/*****************************************************************************
 * Aircraft:new( x,y,z|Vec3, Aircraft::eType, ModelCache.LOAD_SHARED|LOAD_COPY )
 *****************************************************************************/
INTERN int
Aircraft_New( lua_State* L )
{
	const bool            loadCopy     = POP_BOOL( L );
	const Aircraft::eType aircraftType = (Aircraft::eType) POP_INT( L );
	const WorldVertex     pos          = POP_VECTOR3<WorldVertex>( L );

	shptr<Aircraft> aircraft = program::MakeAircraft( aircraftType, pos, loadCopy );

	return NEW_USERDATA_SHPTR<Aircraft>( L, aircraft, BIND_METATABLE_NAME );
}

/*****************************************************************************
 * Lua's garbage collector indirectly invokes this C++ destructor.
 *****************************************************************************/
INTERN int
Aircraft_Destroy( lua_State* L )
{
	return DESTROY_USERDATA_SHPTR<Aircraft>( L, CheckTypesigAircraft );
}

/*****************************************************************************
 * Aircraft:Reset()
 *****************************************************************************/
INTERN int
Aircraft_Reset( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->Reset();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetState( eState )
 *****************************************************************************/
INTERN int
Aircraft_SetState( lua_State* L )
{
	try
	{
		const Aircraft::eState state = (Aircraft::eState) POP_INT( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetState( state );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * eState Aircraft:GetState()
 *****************************************************************************/
INTERN int
Aircraft_GetState( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushnumber( L, aircraft->GetState() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetMatrix( matrix )
 *****************************************************************************/
INTERN int
Aircraft_SetMatrix( lua_State* L )
{
	try
	{
		Matrix m; POP_TABLE_INTO_MATRIX( L, m );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetMatrix( m, Dyna::PLACEMENT );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * table Aircraft:GetMatrix()
 *****************************************************************************/
INTERN int
Aircraft_GetMatrix( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		PUSH_TABLE_FROM_MATRIX( L, aircraft->GetMatrix() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetPosition( x,y,z|Vec3 )
 *****************************************************************************/
INTERN int
Aircraft_SetPosition( lua_State* L )
{
	try
	{
		const WorldVertex pos = POP_VECTOR3<WorldVertex>( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetPosition( pos );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Vec3 Aircraft:GetPosition()
 *****************************************************************************/
INTERN int
Aircraft_GetPosition( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		const WorldVertex pos = aircraft->GetPosition();
		PUSH_VECTOR3( L, pos );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:Rotate( axis, radian )
 *****************************************************************************/
INTERN int
Aircraft_Rotate( lua_State* L )
{
	try
	{
		const fp   radian = POP_FP( L );
		const uint axis   = POP_INT( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->Rotate( axis, radian );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:PhysicalRotate( axis, controlFraction, controlFreq )
 *****************************************************************************/
INTERN int
Aircraft_PhysicalRotate( lua_State* L )
{
	try
	{
		const Milliseconds controlFreq( POP_INT( L ) );
		const fp controlFrac = POP_FP( L );
		const uint axis  = POP_INT( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->PhysicalRotate( axis, controlFrac, controlFreq );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:GetCraftPhysics()
 *****************************************************************************/
INTERN int
Aircraft_GetCraftPhysics( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		return NEW_USERDATA_PTR( L,
                                 &aircraft->GetCraftPhysics(),
                                 LUABIND_METATABLE_STRING("CraftPhysics") );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * fp Aircraft:GetThrottle()
 *****************************************************************************/
INTERN int
Aircraft_GetThrottle( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushnumber( L, aircraft->GetThrottle() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetThrottle( fp )
 *****************************************************************************/
INTERN int
Aircraft_SetThrottle( lua_State* L )
{
	try
	{
		const fp throttle = POP_FP( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetThrottle( throttle );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetCollidable( bool )
 *****************************************************************************/
INTERN int
Aircraft_SetCollidable( lua_State* L )
{
	try
	{
		const bool f = POP_BOOL( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetCollidable( f );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:SetLandingGear( bool )
 * down=true
 *****************************************************************************/
INTERN int
Aircraft_SetLandingGear( lua_State* L )
{
	try
	{
		const bool down = POP_BOOL( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetLandingGear( down );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfLandingGear()
 * down=true
 *****************************************************************************/
INTERN int
Aircraft_IfLandingGear( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfLandingGear() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:ToggleLandingGear()
 *****************************************************************************/
INTERN int
Aircraft_ToggleLandingGear( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->SetLandingGear( not aircraft->IfLandingGear() );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:ToggleBrakes()
 *****************************************************************************/
INTERN int
Aircraft_ToggleBrakes( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->EnableBrakes( not aircraft->IfBrakes() );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfBrakes()
 *****************************************************************************/
INTERN int
Aircraft_IfBrakes( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfBrakes() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfWheelBrakes()
 *****************************************************************************/
INTERN int
Aircraft_IfWheelBrakes( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfWheelBrakes() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfHasPropeller()
 *****************************************************************************/
INTERN int
Aircraft_IfHasPropeller( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfHasPropeller() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfCrash()
 *****************************************************************************/
INTERN int
Aircraft_IfCrash( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfCrash() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * int Aircraft:GetCauseOfCrash()
 *****************************************************************************/
INTERN int
Aircraft_GetCauseOfCrash( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushnumber( L, aircraft->GetCauseOfCrash() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft:EnableCoordinatedTurn( bool )
 *****************************************************************************/
INTERN int
Aircraft_EnableCoordinatedTurn( lua_State* L )
{
	try
	{
		const bool coturn = POP_BOOL( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->EnableCoordinatedTurn( coturn );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfCoordinatedTurnEnabled()
 *****************************************************************************/
INTERN int
Aircraft_IfCoordinatedTurnEnabled( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfCoordinatedTurnEnabled() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool Aircraft:IfCoordinatedTurnExecuting()
 *****************************************************************************/
INTERN int
Aircraft_IfCoordinatedTurnExecuting( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		lua_pushboolean( L, aircraft->IfCoordinatedTurnExecuting() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Aircraft.PlayEngineSound( volume )
 *****************************************************************************/
INTERN int
Aircraft_PlayEngineSound( lua_State* L )
{
	try
	{
		const fp volume = POP_FP( L );
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->PlayEngineSound( volume );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Aircraft.StopEngineSound()
 *****************************************************************************/
INTERN int
Aircraft_StopEngineSound( lua_State* L )
{
	try
	{
		shptr<Aircraft> aircraft = POP_USERDATA_AS_SHPTR<Aircraft>( L );
		CHECK_TYPESIG(aircraft,TYPESIG_AIRCRAFT);
		aircraft->StopEngineSound();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN
INTERN const struct luaL_reg sMethodTableNew[] =
{
	{	"new",			Aircraft_New	},
	{	NULL,			NULL			}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"__gc",					Aircraft_Destroy },
	{	"Reset",				Aircraft_Reset },
	{	"SetState",				Aircraft_SetState },
	{	"GetState",				Aircraft_GetState },
	{	"SetMatrix",			Aircraft_SetMatrix },
	{	"GetMatrix",			Aircraft_GetMatrix },
	{	"SetPosition",			Aircraft_SetPosition },
	{	"GetPosition",			Aircraft_GetPosition },
	{	"Rotate",				Aircraft_Rotate },
	{	"PhysicalRotate",		Aircraft_PhysicalRotate },
	{	"GetCraftPhysics",		Aircraft_GetCraftPhysics },
	{	"SetThrottle",			Aircraft_SetThrottle },
	{	"GetThrottle",			Aircraft_GetThrottle },
	{	"SetCollidable",		Aircraft_SetCollidable },
	{	"SetLandingGear",		Aircraft_SetLandingGear },
	{	"IfLandingGear",		Aircraft_IfLandingGear },
	{	"ToggleLandingGear",	Aircraft_ToggleLandingGear },
	{	"ToggleBrakes",			Aircraft_ToggleBrakes },
	{	"IfHasPropeller",		Aircraft_IfHasPropeller },
	{	"IfBrakes",				Aircraft_IfBrakes },
	{	"IfWheelBrakes",		Aircraft_IfWheelBrakes },
	{	"IfCrash",				Aircraft_IfCrash },
	{	"GetCauseOfCrash",		Aircraft_GetCauseOfCrash },
	{	"EnableCoordinatedTurn",      Aircraft_EnableCoordinatedTurn },
	{	"IfCoordinatedTurnEnabled",   Aircraft_IfCoordinatedTurnEnabled },
	{	"IfCoordinatedTurnExecuting", Aircraft_IfCoordinatedTurnExecuting },
	{	"PlayEngineSound",		Aircraft_PlayEngineSound },
	{	"StopEngineSound",		Aircraft_StopEngineSound },
	{	NULL,					NULL }
};

void
RegisterAircraft( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}
#endif // ! DOXYGEN

} // namespace lua_bind
