/*****************************************************************************
 * $LastChangedDate: 2008-06-13 22:41:29 -0400 (Fri, 13 Jun 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "view/module.hh"
#include "view/view.hh"
#include "view/chase_viewpoint.hh"
using namespace view;
#include "object/module.hh"
using namespace object;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "ChaseViewpoint"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

#if ! DOXYGEN
// Need a function to check typesig because of macro/template limitations.
INTERN void
CheckTypesigChaseViewpoint( shptr<ChaseViewpoint> viewpoint )
{
	CHECK_TYPESIG(viewpoint,TYPESIG_CHASE_VIEWPOINT);
}
#endif // ! DOXYGEN

/*****************************************************************************
 * Viewpoint:new()
 *****************************************************************************/
INTERN int
ChaseViewpoint_New( lua_State* L )
{
	shptr<ChaseViewpoint> viewpoint = new ChaseViewpoint;
	return NEW_USERDATA_SHPTR<ChaseViewpoint>( L, viewpoint, BIND_METATABLE_NAME );
}

/*****************************************************************************
 * Lua's garbage collector indirectly invokes this C++ destructor.
 *****************************************************************************/
INTERN int
ChaseViewpoint_Destroy( lua_State* L )
{
	return DESTROY_USERDATA_SHPTR<ChaseViewpoint>( L, CheckTypesigChaseViewpoint );
}

/*****************************************************************************
 * ChaseViewpoint:Save()
 *****************************************************************************/
INTERN int
ChaseViewpoint_Save( lua_State* L )
{
	try
	{
		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);
		chaseViewpoint->Save();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * ChaseViewpoint:Reset()
 *****************************************************************************/
INTERN int
ChaseViewpoint_Reset( lua_State* L )
{
	try
	{
		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);
		chaseViewpoint->Reset();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * ChaseViewpoint:SetTarget( Object )
 *****************************************************************************/
INTERN int
ChaseViewpoint_SetTarget( lua_State* L )
{
	try
	{
		// Example of Lua passing an Object-derivative (usually Aircraft)
		// which C++ then casts to a base class.
		shptr<Object> target = POP_USERDATA_AS_SHPTR<Object>( L );
		CHECK_TYPESIG(target,TYPESIG_OBJECT);

		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);

		chaseViewpoint->SetTarget( target );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * ChaseViewpoint:SetDistance( x,y,z|Vec3 )
 *****************************************************************************/
INTERN int
ChaseViewpoint_SetDistance( lua_State* L )
{
	try
	{
		const Vector3 distance = POP_VECTOR3<Vector3>( L );
		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);
		chaseViewpoint->SetDistance( distance );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * ChaseViewpoint:Rotate( axis, radian )
 *****************************************************************************/
INTERN int
ChaseViewpoint_Rotate( lua_State* L )
{
	try
	{
		const Radian rad( POP_FP( L ) );
		const uint axis( POP_INT( L ) );
		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);
		chaseViewpoint->Rotate( axis, rad );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * ChaseViewpoint:Move( inc )
 * inc is in world space.
 *****************************************************************************/
INTERN int
ChaseViewpoint_Move( lua_State* L )
{
	try
	{
		const fp inc = POP_FP( L );
		shptr<ChaseViewpoint> chaseViewpoint = POP_USERDATA_AS_SHPTR<ChaseViewpoint>( L );
		CHECK_TYPESIG(chaseViewpoint,TYPESIG_CHASE_VIEWPOINT);
		chaseViewpoint->Move( inc );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN

INTERN const struct luaL_reg sMethodTableNew[] =
{
	{	"new",				ChaseViewpoint_New		},
	{	NULL,				NULL		}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"__gc",				ChaseViewpoint_Destroy },
	{	"Save",				ChaseViewpoint_Save },
	{	"Reset",			ChaseViewpoint_Reset },
	{	"SetTarget",		ChaseViewpoint_SetTarget },
	{	"SetDistance",		ChaseViewpoint_SetDistance },
	{	"Rotate",			ChaseViewpoint_Rotate },
	{	"Move",				ChaseViewpoint_Move },
	{	NULL,				NULL }
};

void
RegisterChaseViewpoint( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}

#endif // ! DOXYGEN

} // namespace lua_bind
