/*****************************************************************************
 * $LastChangedDate: 2008-06-13 22:41:29 -0400 (Fri, 13 Jun 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

// CraftPhysics isn't reference-counted as it is an actual member of Craft.

#include "base/module.hh"
using namespace base;
#include "physics/module.hh"
#include "physics/physics_craft.hh"
using namespace physics;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "CraftPhysics"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

/*****************************************************************************
 * CraftPhysics:Enable( bool )
 *****************************************************************************/
INTERN int
CraftPhysics_Enable( lua_State* L )
{
	try
	{
		const bool enable = POP_BOOL( L );
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		craftPhysics->Enable( enable );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * CraftPhysics:SetSpeed( Speed )
 *****************************************************************************/
INTERN int
CraftPhysics_SetSpeed( lua_State* L )
{
	try
	{
		const Speed speed( POP_NUMBER( L ) );
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		craftPhysics->SetSpeed( speed );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Speed CraftPhysics:ComputeSpeed()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeSpeed( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		lua_pushnumber( L, craftPhysics->ComputeSpeed().FP() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Newton[vec3] CraftPhysics:ComputeThrustForce()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeThrustForce( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		PUSH_VECTOR3( L, craftPhysics->ComputeThrustForce() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Newton[vec3] CraftPhysics:ComputeLiftForce()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeLiftForce( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		PUSH_VECTOR3( L, craftPhysics->ComputeLiftForce() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Newton[vec3] CraftPhysics:ComputeDragForce()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeDragForce( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		PUSH_VECTOR3( L, craftPhysics->ComputeDragForce() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Newton[vec3] CraftPhysics:ComputeWeightForce()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeWeightForce( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		PUSH_VECTOR3( L, craftPhysics->ComputeWeightForce() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Degree CraftPhysics:ComputeAngleOfAttack()
 *****************************************************************************/
INTERN int
CraftPhysics_ComputeAngleOfAttack( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		lua_pushnumber( L, craftPhysics->ComputeAngleOfAttack() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool CraftPhysics:IfStall()
 *****************************************************************************/
INTERN int
CraftPhysics_IfStall( lua_State* L )
{
	try
	{
		SafePtr<CraftPhysics> craftPhysics = POP_USERDATA_AS_PTR<CraftPhysics>( L );
		CHECK_TYPESIG(craftPhysics,TYPESIG_CRAFT_PHYSICS);
		lua_pushboolean( L, craftPhysics->IfStall() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN

INTERN const struct luaL_reg sMethodTableNew[] =
{
	// (new and __gc omitted)
	{	NULL,			NULL		}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"Enable",				CraftPhysics_Enable },
	{	"SetSpeed",				CraftPhysics_SetSpeed },
	{	"ComputeSpeed",			CraftPhysics_ComputeSpeed },
	{	"ComputeThrustForce",	CraftPhysics_ComputeThrustForce },
	{	"ComputeLiftForce",		CraftPhysics_ComputeLiftForce },
	{	"ComputeDragForce",		CraftPhysics_ComputeDragForce },
	{	"ComputeWeightForce",	CraftPhysics_ComputeWeightForce },
	{	"ComputeAngleOfAttack",	CraftPhysics_ComputeAngleOfAttack },
	{	"IfStall",				CraftPhysics_IfStall },
	{	NULL,					NULL }
};

void
RegisterCraftPhysics( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}

#endif // ! DOXYGEN

} // namespace lua_bind
