/*****************************************************************************
 * $LastChangedDate: 2008-06-13 22:41:29 -0400 (Fri, 13 Jun 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "hud/module.hh"
using namespace hud;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "HudString"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

#if ! DOXYGEN

// Need a function to check typesig because of macro/template limitations.
INTERN void
CheckTypesigHudString( shptr<Hud::HudString> obj )
{
	CHECK_TYPESIG(obj,TYPESIG_HUDSTRING);
}

#endif // ! DOXYGEN

/*****************************************************************************
 * HudString:new( { text="", orientation=, x=, y=, fontSize= } )
 * text and orientation are required.
 *****************************************************************************/
INTERN int
HudString_New( lua_State* L )
{
	// Special case: Since some table members are optional,
	// a C++ exception could be thrown during stack operations.
	SAVE_STACK( L );

	// Fill a HudString::Args struct with optional members of a Lua table.
	Hud::HudString::Args hudStringArgs;

	// .mText
	try
	{
		hudStringArgs.mText = GET_MEMBER_AS_STRING( L, "text" );
	}
	CATCH_LUA_EXCEPTION

	// .mOrientation
	try
	{
		const string orientation = GET_MEMBER_AS_STRING( L, "orientation" );
		if ( orientation == "TOP_LEFT" )
			hudStringArgs.mOrientation = Hud::HudString::eOrientation_TOP_LEFT;
		else if ( orientation == "TOP_RIGHT" )
			hudStringArgs.mOrientation = Hud::HudString::eOrientation_TOP_RIGHT;
		else if ( orientation == "BOTTOM_LEFT" )
			hudStringArgs.mOrientation = Hud::HudString::eOrientation_BOTTOM_LEFT;
		else if ( orientation == "BOTTOM_RIGHT" )
			hudStringArgs.mOrientation = Hud::HudString::eOrientation_BOTTOM_RIGHT;
		else if ( orientation == "CENTER" )
			hudStringArgs.mOrientation = Hud::HudString::eOrientation_CENTER;
	}
	CATCH_LUA_EXCEPTION

	// .mPos
	try
	{
		hudStringArgs.mPos[XX] = GET_MEMBER_AS_NUMBER( L, "x", OPTIONAL );
		hudStringArgs.mPos[YY] = GET_MEMBER_AS_NUMBER( L, "y", OPTIONAL );
	}
	catch( const LuaException& ) { } // ignore
	RESTORE_STACK( L );

	// .mFontSize
	try
	{
		hudStringArgs.mFontSize = GET_MEMBER_AS_NUMBER( L, "fontSize", OPTIONAL );
	}
	catch( const LuaException& ) { } // ignore
	RESTORE_STACK( L );

	// Construct HudString from HudString::Args.
	shptr<Hud::HudString> hudString = new Hud::HudString( hudStringArgs );

	// Attach to the Hud.
	Hud::GetInstance().AttachHudString( hudString );

	// Pop the passed table arg.
	lua_pop( L, 1 );

	return NEW_USERDATA_SHPTR<Hud::HudString>( L, hudString, BIND_METATABLE_NAME );
}

/*****************************************************************************
 * Lua's garbage collector indirectly invokes this C++ destructor.
 *****************************************************************************/
INTERN int
HudString_Destroy( lua_State* L )
{
	return DESTROY_USERDATA_SHPTR<Hud::HudString>( L, CheckTypesigHudString );
}

/*****************************************************************************
 * HudString:SetText( string )
 *****************************************************************************/
INTERN int
HudString_SetText( lua_State* L )
{
	try
	{
		const string text = POP_STRING( L );
		shptr<Hud::HudString> hudString = POP_USERDATA_AS_SHPTR<Hud::HudString>( L );
		CHECK_TYPESIG( hudString, TYPESIG_HUDSTRING );
		hudString->SetText( text );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * HudString:SetColor( r, g, b, a )  (fp,fp,fp,fp)
 *****************************************************************************/
INTERN int
HudString_SetColor( lua_State* L )
{
	try
	{
		const RGBA color = POP_RGBA( L );
		shptr<Hud::HudString> hudString = POP_USERDATA_AS_SHPTR<Hud::HudString>( L );
		CHECK_TYPESIG( hudString, TYPESIG_HUDSTRING );
		hudString->SetColor( color );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * HudString:SetPosition3D( x,y,z|Vec3 )
 * Set the position of a HUD string by projecting a 3D coordinate into 2D.
 *****************************************************************************/
INTERN int
HudString_SetPosition3D( lua_State* L )
{
	try
	{
		const WorldVertex pos3 = POP_VECTOR3<WorldVertex>( L );
		shptr<Hud::HudString> hudString = POP_USERDATA_AS_SHPTR<Hud::HudString>( L );
		CHECK_TYPESIG( hudString, TYPESIG_HUDSTRING );
		hudString->SetPosition3D( pos3 );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * HudString:SetFontSize( fp )
 *****************************************************************************/
INTERN int
HudString_SetFontSize( lua_State* L )
{
	try
	{
		const fp fontSize = POP_FP( L );
		shptr<Hud::HudString> hudString = POP_USERDATA_AS_SHPTR<Hud::HudString>( L );
		CHECK_TYPESIG( hudString, TYPESIG_HUDSTRING );
		hudString->SetFontSize( fontSize );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN

INTERN const struct luaL_reg sMethodTableNew[] =
{
	{	"new",			HudString_New	},
	{	NULL,			NULL			}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"__gc",				HudString_Destroy },
	{	"SetText",			HudString_SetText },
	{	"SetColor",			HudString_SetColor },
	{	"SetPosition3D",	HudString_SetPosition3D },
	{	"SetFontSize",		HudString_SetFontSize },
	{	NULL,				NULL }
};

void
RegisterHudString( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}

#endif // ! DOXYGEN

} // namespace lua_bind
