/*****************************************************************************
 * $LastChangedDate: 2008-07-04 22:30:59 -0400 (Fri, 04 Jul 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "object/module.hh"
#include "object/object.hh"
using namespace object;
#include "graph/module.hh"
#include "graph/model_cache.hh"
using namespace graph;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "Object"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

#if ! DOXYGEN
// Need a function to check typesig because of macro/template limitations.
INTERN void
CheckTypesigObject( shptr<Object> obj )
{
	CHECK_TYPESIG(obj,TYPESIG_OBJECT);
}
#endif // ! DOXYGEN

/*****************************************************************************
 * Object:new( x,y,z|Vec3, modelPathname, scale, ModelCache.LOAD_SHARED|LOAD_COPY )
 *****************************************************************************/
INTERN int
Object_New( lua_State* L )
{
	const bool        loadCopy  = POP_BOOL( L );
	const fp          scale     = POP_FP( L );
	const string      modelName = POP_STRING( L );
	const WorldVertex pos       = POP_VECTOR3<WorldVertex>( L );

	// Load 3D model.
	shptr<Graph> graph = ModelCache::GetInstance().LoadModel( modelName, scale, loadCopy );
	shptr<Object> object = new Object( graph, pos );

	return NEW_USERDATA_SHPTR<Object>( L, object, BIND_METATABLE_NAME );
}

/*****************************************************************************
 * Lua's garbage collector indirectly invokes this C++ destructor.
 *****************************************************************************/
INTERN int
Object_Destroy( lua_State* L )
{
	return DESTROY_USERDATA_SHPTR<Object>( L, CheckTypesigObject );
}

/*****************************************************************************
 * Object:SetPosition( x,y,z|Vec3 )
 *****************************************************************************/
INTERN int
Object_SetPosition( lua_State* L )
{
	try
	{
		const WorldVertex pos = POP_VECTOR3<WorldVertex>( L );
		shptr<Object> object = POP_USERDATA_AS_SHPTR<Object>( L );
		CHECK_TYPESIG(object,TYPESIG_OBJECT);
		object->SetPosition( pos );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Vec3 Object:GetPosition()
 *****************************************************************************/
INTERN int
Object_GetPosition( lua_State* L )
{
	try
	{
		shptr<Object> object = POP_USERDATA_AS_SHPTR<Object>( L );
		CHECK_TYPESIG(object,TYPESIG_OBJECT);
		const WorldVertex pos = object->GetPosition();
		PUSH_VECTOR3( L, pos );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Object:Rotate( axis, radian )
 *****************************************************************************/
INTERN int
Object_Rotate( lua_State* L )
{
	try
	{
		const fp   radian = POP_FP( L );
		const uint axis   = POP_INT( L );
		shptr<Object> object = POP_USERDATA_AS_SHPTR<Object>( L );
		CHECK_TYPESIG(object,TYPESIG_OBJECT);
		object->Rotate( axis, radian );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Object:SetCollidable( bool )
 *****************************************************************************/
INTERN int
Object_SetCollidable( lua_State* L )
{
	try
	{
		const bool f = POP_BOOL( L );
		shptr<Object> object = POP_USERDATA_AS_SHPTR<Object>( L );
		CHECK_TYPESIG(object,TYPESIG_OBJECT);
		object->SetCollidable( f );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN
INTERN const struct luaL_reg sMethodTableNew[] =
{
	{	"new",					Object_New	},
	{	NULL,					NULL		}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"__gc",					Object_Destroy },
	{	"SetPosition",			Object_SetPosition },
	{	"GetPosition",			Object_GetPosition },
	{	"Rotate",				Object_Rotate },
	{	"SetCollidable",		Object_SetCollidable },
	{	NULL,					NULL }
};

void
RegisterObject( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}
#endif // ! DOXYGEN

} // namespace lua_bind
