/*****************************************************************************
 * $LastChangedDate: 2008-09-20 10:29:31 -0400 (Sat, 20 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "program/module.hh"
using namespace program;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "Program"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

/*****************************************************************************
 * string Program:GetScriptDir()
 *****************************************************************************/
INTERN int
Program_GetScriptDir( lua_State* L )
{
	lua_pushstring( L, GET_LUA().GetScriptDir().c_str() );
	return 1;  // number of pushed values
}

/*****************************************************************************
 * milliseconds Program:GetElapsedTime()
 *****************************************************************************/
INTERN int
Program_GetElapsedTime( lua_State* L )
{
	lua_pushnumber( L, base::Timer::GetInstance().GetElapsedTime() );
	return 1;  // number of pushed values
}

/*****************************************************************************
 * Program:Reset()
 *****************************************************************************/
INTERN int
Program_Reset( lua_State* L )
{
	try
	{
		gProgram.Reset();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Program:Exit( int )
 *****************************************************************************/
INTERN int
Program_Exit( lua_State* L )
{
	try
	{
		EXIT( POP_INT( L ) );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Program:SetPause( bool )
 *****************************************************************************/
INTERN int
Program_SetPause( lua_State* L )
{
	try
	{
		gProgram.SetPause( POP_BOOL( L ) );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Program:IfPaused()
 *****************************************************************************/
INTERN int
Program_IfPaused( lua_State* L )
{
	try
	{
		lua_pushboolean( L, gProgram.IfPaused() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * milliseconds Program:GetTickFreq()
 * Frequency of program's timer-tick (its pulse).
 *****************************************************************************/
INTERN int
Program_GetTickFreq( lua_State* L )
{
	try
	{
		lua_pushnumber( L, PROGRAM_CONF.mTimerTickFreq );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

/*****************************************************************************
 * string Program:DequeueCommand()
 * Returns next command or nil.
 *****************************************************************************/
INTERN int
Program_DequeueCommand( lua_State* L )
{
	try
	{
		const string cmd = gProgram.DequeueCommand();
		if ( cmd.empty() )
			lua_pushnil( L );
		else
			lua_pushstring( L, cmd.c_str() );
	}
	CATCH_LUA_EXCEPTION
	return 1;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN

INTERN const struct luaL_reg sMethodTableNew[] =
{
	// (new and __gc omitted)
	{	NULL,			NULL		}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"Reset",				Program_Reset },
	{	"Exit",					Program_Exit },
	{	"SetPause",				Program_SetPause },
	{	"IfPaused",				Program_IfPaused },
	{	"GetScriptDir",			Program_GetScriptDir },
	{	"GetElapsedTime",		Program_GetElapsedTime },
	{	"GetTickFreq",			Program_GetTickFreq },
	{	"DequeueCommand",		Program_DequeueCommand },
	{	NULL,					NULL }
};

void
RegisterProgram( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}

#endif // ! DOXYGEN

} // namespace lua_bind
