/*****************************************************************************
 * $LastChangedDate: 2008-09-23 11:09:26 -0400 (Tue, 23 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Interface of C functions for Lua scripts.
 * @remarks module.txt documents how to export new C++ classes to Lua.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#include "base/module.hh"
using namespace base;
#include "control/replay.hh"
using namespace control;
#include "lua_bind/module.hh"
#include "lua_bind/lua_lang.hh"
#include "lua_bind/bind.hh"

namespace lua_bind {

#define BIND_CLASS_NAME      "Replay"
#define BIND_METATABLE_NAME  LUABIND_METATABLE_STRING(BIND_CLASS_NAME)

//==============================================================================
// C functions called by Lua that access C++ objects.
//==============================================================================

#if ! DOXYGEN
// Need a function to check typesig because of macro/template limitations.
INTERN void
CheckTypesigReplay( shptr<Replay> obj )
{
	CHECK_TYPESIG(obj,TYPESIG_REPLAY);
}
#endif // ! DOXYGEN

/*****************************************************************************
 * Replay:new( craft, milliseconds, commandMode )
 *****************************************************************************/
INTERN int
Replay_New( lua_State* L )
{
	try
	{
		const Command::ExecuteMode commandExecuteMode = Command::ExecuteMode( POP_INT( L ) );
		const Milliseconds milliseconds = POP_INT( L );
		shptr<Craft> craft = POP_USERDATA_AS_SHPTR<Craft>( L );
		CHECK_TYPESIG(craft,TYPESIG_CRAFT);
		shptr<Replay> replay = new Replay( craft, milliseconds, commandExecuteMode );
		return NEW_USERDATA_SHPTR<Replay>( L, replay, BIND_METATABLE_NAME );
	}
	CATCH_LUA_EXCEPTION
	return 0;  // oops
}

/*****************************************************************************
 * Lua's garbage collector indirectly invokes this C++ destructor.
 *****************************************************************************/
INTERN int
Replay_Destroy( lua_State* L )
{
	return DESTROY_USERDATA_SHPTR<Replay>( L, CheckTypesigReplay );
}

/*****************************************************************************
 * Replay:Record()
 *****************************************************************************/
INTERN int
Replay_Record( lua_State* L )
{
	try
	{
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		replay->Record();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Replay:Play()
 *****************************************************************************/
INTERN int
Replay_Play( lua_State* L )
{
	try
	{
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		replay->Play();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * Replay:Stop()
 * Stop playing or recording (whichever is active).
 *****************************************************************************/
INTERN int
Replay_Stop( lua_State* L )
{
	try
	{
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		replay->Stop();
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Replay:IfPlaying()
 *****************************************************************************/
INTERN int
Replay_IfPlaying( lua_State* L )
{
	try
	{
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		lua_pushboolean( L, replay->IfPlaying() );
		return 1;
	}
	CATCH_LUA_EXCEPTION
	return 0;  // number of pushed values
}

/*****************************************************************************
 * bool Replay:Save( filename )
 *****************************************************************************/
INTERN int
Replay_Save( lua_State* L )
{
	try
	{
		const string filename = POP_STRING( L );
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		try
		{
			const bool saved = replay->Save( filename );
			lua_pushboolean( L, saved );
			return 1;
		}
		catch( const Exception& e )
		{
			CERROR << e.GetText() << endl;
			lua_pushboolean( L, false );
			return 1;
		}
	}
	CATCH_LUA_EXCEPTION
	lua_pushboolean( L, false );
	return 1;  // number of pushed values
}

/*****************************************************************************
 * bool Replay:Load( filename )
 *****************************************************************************/
INTERN int
Replay_Load( lua_State* L )
{
	try
	{
		const string filename = POP_STRING( L );
		shptr<Replay> replay = POP_USERDATA_AS_SHPTR<Replay>( L );
		CHECK_TYPESIG(replay,TYPESIG_REPLAY);
		// Load() will throw exception if file doesn't exist.
		try
		{
			const bool loaded = replay->Load( filename );
			lua_pushboolean( L, loaded );
			return 1;
		}
		catch( const Exception& e )
		{
			CERROR << e.GetText() << endl;
			lua_pushboolean( L, false );
			return 1;
		}
	}
	CATCH_LUA_EXCEPTION
	lua_pushboolean( L, false );
	return 1;  // number of pushed values
}

//==============================================================================
// Register C functions with Lua.
//==============================================================================

#if ! DOXYGEN
INTERN const struct luaL_reg sMethodTableNew[] =
{
	{	"new",			Replay_New	},
	{	NULL,			NULL		}
};
    
INTERN const struct luaL_reg sMethodTable[] =
{
	{	"__gc",				Replay_Destroy },
	{	"Record",			Replay_Record },
	{	"Play",				Replay_Play },
	{	"Stop",				Replay_Stop },
	{	"IfPlaying",		Replay_IfPlaying },
	{	"Save",				Replay_Save },
	{	"Load",				Replay_Load },
	{	NULL,				NULL }
};

void
RegisterReplay( lua_State* L )
{
	REGISTER_CLASS_METATABLE( L,
                              BIND_METATABLE_NAME,
                              BIND_CLASS_NAME,
                              sMethodTableNew,
                              sMethodTable );
}
#endif // ! DOXYGEN

} // namespace lua_bind
