/*****************************************************************************
 * $LastChangedDate: 2008-09-14 00:08:05 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Craft class (base class for aircraft).
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define OBJECT_CRAFT_CC 1
#include "base/module.hh"
#include "base/random.hh"
using namespace base;
#include "view/module.hh"
using namespace view;
#include "physics/module.hh"
#include "physics/physics_craft.hh"
using namespace physics;
#include "control/module.hh"
#include "control/conf.hh"
#include "control/command.hh"
#include "control/events.hh"
using namespace control;
#include "world/module.hh"
using namespace world;
#include "object/module.hh"
#include "object/defs.hh"
#include "object/craft.hh"

namespace object {

const fp			MANEUVER_LIFTOFF_DELTA = 20.f;
const Degree		STALL_DEGREE_ANIMATION( 0.1f );
const Milliseconds	STATE_UPDATE_FREQ( 500 );

// Abbrev.
#define CRAFT_GEO_POSITION() \
	(Geo::WorldVertex2GeoVertex( GetPosition() ))

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  Craft  //////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Craft::Craft( shptr<Graph> graph, const WorldVertex& pos, const CraftSpecs& craftSpecs )
:	Parent(graph,pos),
	mCraftPhysics(this,craftSpecs),
    mState(eState_PARKED),
	mStateUpdateTime(0),
	mCauseOfCrash(eCauseOfCrash_DEFAULT),
	mCoordinatedTurnEnabled(false),
	mCoordinatedTurnExecuting(false),
	mThrottle(0.0f),
	mLandingGearDown(false)
{
	SET_TYPESIG(this,TYPESIG_CRAFT);
}

// LIMITED USE
Craft::Craft( void )
:	Parent(Graph::MakeMinimalGraph(),WorldVertex(0,0,0)),
	mCraftPhysics(this,CraftSpecs()),
    mState(eState_PARKED),
	mStateUpdateTime(0),
	mCauseOfCrash(eCauseOfCrash_DEFAULT),
	mCoordinatedTurnEnabled(false),
	mCoordinatedTurnExecuting(false),
	mThrottle(0.0f),
	mLandingGearDown(false)
{
	SET_TYPESIG(this,TYPESIG_CRAFT);
}

Craft::~Craft()
{
	INVALIDATE_TYPESIG(this,TYPESIG_CRAFT);
}

/*****************************************************************************
 * Alternative to a NULL Craft.
 *****************************************************************************/
shptr<Craft>
Craft::MakeDummyCraft( void )
{
	return new Craft;
}

/*****************************************************************************
 * Reset.
 *****************************************************************************/
void
Craft::Reset( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::Reset();

	mState        = eState_PARKED;  // ok, derived Reset() can reassign
	mCauseOfCrash = eCauseOfCrash_DEFAULT;
	mThrottle     = 0.0f;

	// Reset physics (stop movement).
	mCraftPhysics.Reset();
}

/*****************************************************************************
 * Set position of Craft.
 *****************************************************************************/
void
Craft::SetPosition( const WorldVertex& position )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::SetPosition( position );
}

/*****************************************************************************
 * Translate matrix.
 *****************************************************************************/
void
Craft::Translate( uint axis, fp inc )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::Translate( axis, inc );
}

void
Craft::Translate( const Vector3& v )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::Translate( v );
}

/*****************************************************************************
 * Prevent rotation while aircraft is in contact with runway.
 *****************************************************************************/
void
Craft::Rotate( uint axis, Radian rad )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	// Prevent rotation while aircraft is in contact with runway.
	if ( not Geo::IfContactForce( CRAFT_GEO_POSITION() )
     or mCraftPhysics.ComputeVelocity()[ZZ] > 0.25f )  // or lifting
	{
		Parent::Rotate( axis, rad );
	}
}

/*****************************************************************************
 * Set matrix.
 *****************************************************************************/
void
Craft::SetMatrix( const Matrix& matrix, Dyna::eMatrixChange change )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::SetMatrix( matrix, change );

	// Broadcast matrix as separate quaternion and position.
	// Replay can decide which is a change (avoid recording repeated commands).
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandQuaternion( Quaternion(matrix) ) ) );
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandPosition( CRAFT_GEO_POSITION() ) ) );
}

/*****************************************************************************
 * Rotate by a degree according to physics (airframe), amount of control input, and time.
 *
 * This is called by the joystick/keyboard handlers that control the "current" Craft.
 *
 * For example, if the user pulls all the way left on a joystick
 * which is polled every 20 milliseconds:
 * PhysicalRotate( AXIS_ROLL, -1.0f, Milliseconds(20) )
 * This would rotate the aircraft at its maximum rate-of-roll.
 * The rate-of-roll (degree/second) would be divided down for 20 milliseconds.
 *
 * @param	axis
 * @param	controlFraction
 *          {-1.0,..,1.0}
 * @param	controlFreq
 *          The polling frequency of joystick/keyboard input.
 *****************************************************************************/
void
Craft::PhysicalRotate( uint axis, fp controlFraction, Milliseconds controlFreq )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);
ASSERT_AXIS( axis );
ASSERT( ABS(controlFraction) < 1.1f );
ASSERT( uint(controlFreq) > 0 );
ASSERT( uint(controlFreq) < 100 );

	// Don't rotate if crashed.
	if ( IfCrash() )
		return;

	// Don't rotate at low speeds on runway.
	const Speed manueveringSpeed = \
		GetCraftPhysics().ComputeLiftoffSpeed() - KPH2Speed(SpeedKPH(MANEUVER_LIFTOFF_DELTA));
	if ( IfCollisionRunway() and (GetCraftPhysics().ComputeSpeed() < manueveringSpeed ) )
		return;

	// Get the maximum rotation rate (degree/sec) of this axis.
	Degree deg = 0.0f;
	switch ( axis )
	{
		case AXIS_ROLL:  deg = mCraftPhysics.ComputeRollRate();  break;
		case AXIS_PITCH: deg = mCraftPhysics.ComputePitchRate(); break;
		case AXIS_YAW:   deg = mCraftPhysics.ComputeYawRate();   break;
	}
	ASSERT( FP_GE<fp>(deg,0.0f) );  // might be zero degree

	// Multiply by the amount of control input.
	// Pitching up/down all-the-way: fraction = +-1.0
	deg = deg * controlFraction;

	// Divide since rate is measured in seconds
	// but this rotation occurs in milliseconds.
	deg = deg * (fp(controlFreq) / 1000.f);

	// Directly rotate.
	const Radian rad = Deg2Rad( deg );
	Rotate( axis, rad );

	// Coordinated turn:
	// Match yaw with pitch (absolute, not physical/adjusted) during a coordinated turn.
	// Only for pitch up, not pitch down, as pitch up tightens the turns.
	mCoordinatedTurnExecuting = false;
	if ( mCoordinatedTurnEnabled
     and axis == AXIS_PITCH
     and IfSameSign<fp>(deg,Control::GetInstance().GetPitchUpDir()) )
	{
		CraftPhysics::eTurning turning = mCraftPhysics.ComputeTurningDir();
		const Radian radYaw = fp(rad) * 1.05f;
		if ( turning == CraftPhysics::eTurningLeft )
		{
			mCoordinatedTurnExecuting = true;
			Rotate( AXIS_YAW, ABS(radYaw) * Control::GetInstance().GetYawLeftDir() );
		}
        else if ( turning == CraftPhysics::eTurningRight )
		{
			mCoordinatedTurnExecuting = true;
			Rotate( AXIS_YAW, ABS(radYaw) * Control::GetInstance().GetYawRightDir() );
		}
	}

	// Broadcast command was executed.
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandPhysicalRotate(axis,controlFraction,controlFreq) ) );
}

/*****************************************************************************
 * Set throttle.
 * throttle = {0.0,..,1.0}
 * @param	throttle
 *			This method will clamp if necessary.
 *****************************************************************************/
void
Craft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	mThrottle = Clamp1( throttle );

	// Set thrust of physics model.
	mCraftPhysics.SetThrust( mCraftPhysics.ComputeMaxThrust() * Newton1(mThrottle) );

	// Broadcast command was executed.
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandThrottle(throttle) ) );
}

/*****************************************************************************
 * Enable/disable brakes.
 *****************************************************************************/
void
Craft::EnableBrakes( const bool enable )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	// CraftPhysics determines magnitude of drag force
	// depending on air-brakes or wheel-brakes.
	GetCraftPhysics().EnableBrakes( enable );

	// Broadcast command was executed.
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandBrakes(enable) ) );
}

bool
Craft::IfBrakes( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return GetCraftPhysics().IfBrakes();
}

bool
Craft::IfWheelBrakes( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return GetCraftPhysics().IfWheelBrakes();
}

bool
Craft::IfAirBrakes( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return GetCraftPhysics().IfAirBrakes();
}

/*****************************************************************************
 * Enable/disable landing gear.
 *****************************************************************************/
void
Craft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	mLandingGearDown = down;

	// Broadcast command was executed.
	CONTROL_EVENTS.mCommand.Broadcast( CommandEvent( this, new CommandLandingGear(down) ) );
}

/*****************************************************************************
 * @return True if landing gear is extended down.
 *****************************************************************************/
bool
Craft::IfLandingGear( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return mLandingGearDown;
}

/*****************************************************************************
 * Coordinated-turn methods.
 *****************************************************************************/
void
Craft::EnableCoordinatedTurn( const bool enable )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	mCoordinatedTurnEnabled = enable;
}

bool
Craft::IfCoordinatedTurnEnabled( void ) const
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return mCoordinatedTurnEnabled;
}

bool
Craft::IfCoordinatedTurnExecuting( void ) const
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	return mCoordinatedTurnExecuting;
}

/*****************************************************************************
 * Animation.
 *****************************************************************************/
void
Craft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	Parent::Tick( millisecElapsed );

    // This updates some states that aren't updated by CraftPhysics.
    UpdateState( millisecElapsed );

	// Physics model will move (and possibly rotate) this Craft.
	mCraftPhysics.Tick( millisecElapsed );

	// Stalling?
	if ( mCraftPhysics.IfStall() )
	{
		AnimateStall();
	}
}

/*****************************************************************************
 * Update Craft for states that aren't computed elsewhere.
 *****************************************************************************/
void
Craft::UpdateState( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	// Dampen state-transitions by updating at a low frequency (a few seconds).
	// A reason is to avoid fluttering between TAKEOFF and FLYING.
	if ( millisecElapsed - mStateUpdateTime < STATE_UPDATE_FREQ )
		return;

	// Update timestamp.
	mStateUpdateTime = millisecElapsed;

	// Vars.
	CraftPhysics& craftPhysics = GetCraftPhysics();
	const Speed   speed        = craftPhysics.ComputeSpeed();
    const bool    onRunway     = IfCollisionRunway(); 

	// State transitions.
	switch ( GetState() )
	{
		case eState_PARKED:
		{
			// PARKED --> TAKEOFF
			if ( speed > SPEED_MINIMAL )
				SetState( eState_TAKEOFF );
		}
		break;

		case eState_TAKEOFF:
		{
			// TAKEOFF --> FLYING
			// TAKEOFF --> PARKED (rare, if pilot stopped during takeoff)
	        if ( not onRunway )
				SetState( eState_FLYING );
			else if ( speed < SPEED_MINIMAL )
				SetState( eState_PARKED );
		}
		break;

		case eState_FLYING:
		{
			// FLYING --> LANDING
	        if ( onRunway )
				SetState( eState_LANDING );
		}
		break;

		case eState_LANDING:
		{
			// LANDING --> FLYING (pilot missed landing on carrier runway)
			// LANDING --> PARKED
			if ( not onRunway )
				SetState( eState_FLYING );
	        else if ( speed < SPEED_MINIMAL )
				SetState( eState_PARKED );
		}
		break;

		default:  // NOP
		break;
	}
}

/*****************************************************************************
 * Animate the aerodynamic stalling of a Craft.
 *****************************************************************************/
void
Craft::AnimateStall( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	// Craft should begin to slightly spin when slight below stall-speed,
	// then gradually spin worse as speed declines.
	// Conversely, spinning should decrease as speed increases.
	// CraftPhysics model will compute 1.0 for maximum stall (or more if excessive stall).
	fp stallFactor = mCraftPhysics.ComputeStall();
	stallFactor = std::sin( stallFactor * math::RADIAN_90 );
	stallFactor += gRandom.random_f( 0.35f );
	Rotate( AXIS_PITCH,  stallFactor * Deg2Rad(STALL_DEGREE_ANIMATION * Degree(3.00f)) * GET_CONTROL().GetPitchDownDir() );
	Rotate( AXIS_YAW,    stallFactor * Deg2Rad(STALL_DEGREE_ANIMATION * Degree(1.00f)) );
	Rotate( AXIS_ROLL,  -stallFactor * Deg2Rad(STALL_DEGREE_ANIMATION * Degree(2.00f)) );
}

/*****************************************************************************
 * Craft states (parked, flying, etc).
 *****************************************************************************/
void
Craft::SetState( const eState state )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

    mState = state;
}

Craft::eState
Craft::GetState( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

    return mState;
}

/*****************************************************************************
 * Crash and collision methods.
 *****************************************************************************/
bool
Craft::IfCrash( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

    return mState == eState_CRASHED;
}

void
Craft::SetCauseOfCrash( const eCauseOfCrash causeOfCrash )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

	mCauseOfCrash = causeOfCrash;
}

Craft::eCauseOfCrash
Craft::GetCauseOfCrash( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);
ASSERT( mState == eState_CRASHED );  // was there a crash?

    return mCauseOfCrash;
}

bool
Craft::IfCollisionRunway( void )
{
CHECK_TYPESIG(this,TYPESIG_CRAFT);

    return IfCollision();  // doesn't really distinguish
}

} // namespace object
