/*****************************************************************************
 * $LastChangedDate: 2008-09-14 00:08:05 -0400 (Sun, 14 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Craft class (base class for aircraft).
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef OBJECT_CRAFT_HH
#define OBJECT_CRAFT_HH 1

#include "object/dyna.hh"
#include "physics/module.hh"
#include "physics/physics_craft.hh"
using namespace physics;

namespace object {

////////////////////////////////////////////////////////////////////////////////
/// @brief Craft class.
///
/// The Craft class is, in fact, a base class specialized for aircraft
/// (Craft's long name would be BaseAircraft).
/// A derived Aircraft class is necessary for app-level actions
/// such as determining if a collision is safe or fatal.
/// This base class doesn't know about other Objects such as runways or carriers,
/// which an Aircraft class in the program module would know.
///
/// mThrottle is a fp {0.0,...,1.0}.
/// mStep is a world coordinate.
/// Speed (translation increment) is determined by mThrottle * mStep.
///
/// Craft state enums:
/// The state enums (flying, landing, etc) are kept in the Craft class.
/// Both Craft and CraftPhysics can affect the state.
///
class Craft : public Dyna
{
PREVENT_COPYING(Craft)
typedef Dyna Parent;
public:                         enum eState
                                {
                                    eState_PARKED,
                                    eState_TAKEOFF,
                                    eState_FLYING,
                                    eState_LANDING,
                                    eState_CRASHED,  // eCauseOfCrash complements this
                                };
public:                         enum eCauseOfCrash
                                {
                                    eCauseOfCrash_DEFAULT,
                                    eCauseOfCrash_DOWNWARD_TOO_FAST
                                };
protected:                      Craft( void );  // LIMITED USE
public:							Craft( shptr<Graph> graph, const WorldVertex& pos, const CraftSpecs& craftSpecs );
public:	   virtual				~Craft();
public:    CLASS_METHOD shptr<Craft> MakeDummyCraft( void );
public:    virtual void         Reset( void );
public:	   virtual const string	GetName( void ) { return string("Craft"); }

// Spatial:
public:    virtual void         SetPosition( const WorldVertex& position );
public:    virtual void			Translate( uint axis, fp inc );
public:    virtual void			Translate( const Vector3& v );
public:	   virtual void			Rotate( uint axis, Radian rad );

// Transformation:
public:    virtual void			SetMatrix( const Matrix& matrix, Dyna::eMatrixChange change );

// Control:
public:	   virtual void			PhysicalRotate( uint axis, fp controlFraction, Milliseconds controlFreq );
public:	   virtual void			SetThrottle( const fp throttle );
public:	   virtual fp			GetThrottle( void ) { return mThrottle; }
public:	   virtual void			EnableBrakes( const bool enable );
public:    virtual bool			IfBrakes( void );
public:    virtual bool			IfWheelBrakes( void );
public:    virtual bool			IfAirBrakes( void );
public:    virtual void			SetLandingGear( const bool down );
public:    virtual bool			IfLandingGear( void );  // true if extended down

// CraftPhysics:
public:	   virtual CraftPhysics& GetCraftPhysics( void ) { CHECK_TYPESIG(this,TYPESIG_CRAFT);
                                                           return mCraftPhysics; }
public:    void					EnableCoordinatedTurn( const bool enable );
public:    bool					IfCoordinatedTurnEnabled( void ) const;
public:    bool					IfCoordinatedTurnExecuting( void ) const;

// Animation:
protected: virtual void			Tick( const Milliseconds millisecElapsed );
protected: virtual void         AnimateStall( void );

// States: collision-detection, crashing, contact force.
// Dyna has basic collision-detection.
// Craft determines what kind of collision: safe contact or fatal collision.
public:    virtual void			SetState( const eState state );
public:    virtual eState       GetState( void );
private:   void                 UpdateState( const Milliseconds millisecElapsed );  // subroutine of Tick()
public:    virtual bool			IfCrash( void );
public:	   virtual void			SetCauseOfCrash( const eCauseOfCrash causeOfCrash );
public:	   virtual eCauseOfCrash GetCauseOfCrash( void );
public:    virtual bool			IfCollisionRunway( void );

// Data:
public:	   CraftPhysics		mCraftPhysics;				///< physics model (public to use CraftPhysics interface)
private:   eState           mState;						///< general states { parked, landing, etc }
private:   Milliseconds     mStateUpdateTime;			///< timestamp when state was updated
private:   eCauseOfCrash	mCauseOfCrash;				///< the cause of the crash (fatal collision)
private:   bool				mCoordinatedTurnEnabled;	///< if coordinated turn is enabled
private:   bool				mCoordinatedTurnExecuting;	///< if program is performing a coordinated turn
private:   fp				mThrottle;					///< {0.0,..,1.0}
protected: bool				mLandingGearDown;
public:	   DECL_TYPESIG( TYPESIG_CRAFT )
};

} // namespace object

#endif // OBJECT_CRAFT_HH
