/*****************************************************************************
 * $LastChangedDate: 2008-03-03 01:07:22 -0500 (Mon, 03 Mar 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Craft specifications.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PHYSICS_CRAFT_SPECS_CC 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "object/module.hh"
#include "object/rotation_list.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "physics/module.hh"
#include "physics/craft_specs.hh"

namespace physics {

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  CraftSpecs  ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
/// @brief Fills a CraftSpecs object from a .conf textual file.
///
/// Mass, max thrust, etc are defined in a data file (physics_*.conf).
/// These parameters are loaded into a CraftSpecs for use by CraftPhysics.
/// CraftSpecs is data, CraftPhysics is computational.
///
class CraftSpecsConfTxt : public base::ConfTxt
{
public:
			CraftSpecsConfTxt( CraftSpecs& craftSpecs ) : mCraftSpecs(craftSpecs) { }
    void	ProcessField( string& field, ifstream& conf );

private:
	CraftSpecs& mCraftSpecs;
};

/*****************************************************************************
 * ctor.
 *****************************************************************************/
CraftSpecs::CraftSpecs( void )
:	mMass(0),			// zero them for assertion checking
	mMaxThrustMag(0),	// magnitude (Newton1)
	mEngineType(eEngineType_PROPELLER),
	mDragAofALimit(0),
	mDragAofADegree(0),
	mLiftoffSpeed(0),
	mMaxAofA(0),
	mRangeOfLiftFactor(0),
	mRangeOfLiftFactorSpeed(0),
	mRangeOfLiftFactorAofA(0),
	mTurnDegreeLow(0),
	mTurnDegreeHigh(0),
	mLength(0),
	mHeight(0),
	mMaxSpeed(0),
	mStallSpeed(0),
	mRollRate(0),
	mPitchRate(0),
	mYawRate(0),
	mModelFile(),
	mModelScale(1.0f),
	mModelOffset(Meter(0),Meter(0),Meter(0)),
	mModelOffsetGear(0),
	mModelRotationList(),
	mModelLighting(true)
{
	// NOP
}

/*****************************************************************************
 * Read a Craft's specificiations/characteristics from a .conf file.
 *****************************************************************************/
void
CraftSpecs::ReadSpecsFromConfFile( const string& confTxtFilename )
{
	try
	{
		CraftSpecsConfTxt confTxt( *this );  // will load this CraftSpecs
		confTxt.Read( string(base::GetDataDir()) + string(confTxtFilename) );
	}
	catch (...)
	{
		CERROR << "ERROR: cannot read " << confTxtFilename << endl;
		EXIT( 1 );
	}
}

void
CraftSpecsConfTxt::ProcessField( string& field, ifstream& conf )
{
	bool error = false;

	// -------------------------- mass -----------------------------------------
	if ( field == "Mass" )
	{
		fp f; conf >> f;
		mCraftSpecs.mMass = Kilogram( f );
	}
	// -------------------------- thrust -----------------------------------------
	else if ( field == "MaxThrust" )
	{
		fp f; conf >> f;
		mCraftSpecs.mMaxThrustMag = Newton1(f);  // magnitude
	}
	else if ( field == "EngineType" )
	{
		// Convert string to enum.
		string s; conf >> s;
		if ( s == "propeller" )
			mCraftSpecs.mEngineType = CraftSpecs::eEngineType_PROPELLER;
		else if ( s == "jet" )
			mCraftSpecs.mEngineType = CraftSpecs::eEngineType_JET;
		else
			error = true;
	}
	// -------------------------- drag -----------------------------------------
	else if ( field == "DragAofALimit" )
	{
		fp f; conf >> f;
		mCraftSpecs.mDragAofALimit = f;
	}
	else if ( field == "DragAofADegree" )
	{
		fp f; conf >> f;
		mCraftSpecs.mDragAofADegree = f;
	}
	// -------------------------- lift -----------------------------------------
	else if ( field == "LiftoffSpeed" )
	{
		fp f; conf >> f;
		mCraftSpecs.mLiftoffSpeed = SpeedKPH(f);
	}
	else if ( field == "MaxAofA" )
	{
		fp f; conf >> f;
		mCraftSpecs.mMaxAofA = f;
	}
	else if ( field == "RangeOfLiftFactor" )
	{
		conf >> mCraftSpecs.mRangeOfLiftFactor;
	}
	else if ( field == "RangeOfLiftFactorSpeed" )
	{
		conf >> mCraftSpecs.mRangeOfLiftFactorSpeed;
	}
	else if ( field == "RangeOfLiftFactorAofA" )
	{
		conf >> mCraftSpecs.mRangeOfLiftFactorAofA;
	}
	// -------------------------- turning -----------------------------------------
	else if ( field == "TurnDegreeLow" )
	{
		fp f; conf >> f;
		mCraftSpecs.mTurnDegreeLow = f;
	}
	else if ( field == "TurnDegreeHigh" )
	{
		fp f; conf >> f;
		mCraftSpecs.mTurnDegreeHigh = f;
	}
	// -------------------------- airframe -----------------------------------------
	else if ( field == "Length" )
	{
		fp f; conf >> f;
		mCraftSpecs.mLength = Meter(f);
	}
	else if ( field == "Height" )
	{
		fp f; conf >> f;
		mCraftSpecs.mHeight = Meter(f);
	}
	else if ( field == "MaxSpeed" )
	{
		fp f; conf >> f;
		mCraftSpecs.mMaxSpeed = SpeedKPH(f);
	}
	else if ( field == "StallSpeed" )
	{
		fp f; conf >> f;
		mCraftSpecs.mStallSpeed = SpeedKPH(f);
	}
	else if ( field == "RollRate" )
	{
		fp f; conf >> f;
		mCraftSpecs.mRollRate = f;
	}
	else if ( field == "PitchRate" )
	{
		fp f; conf >> f;
		mCraftSpecs.mPitchRate = f;
	}
	else if ( field == "YawRate" )
	{
		fp f; conf >> f;
		mCraftSpecs.mYawRate = f;
	}
	// -------------------------- 3D model -------------------------------------
	else if ( field == "ModelFile" )
	{
		conf >> mCraftSpecs.mModelFile;
	}
	else if ( field == "ModelScale" )
	{
		conf >> mCraftSpecs.mModelScale;
	}
	else if ( field == "ModelOffset" )
	{
		fp f;
		conf >> f; mCraftSpecs.mModelOffset[XX] = Meter(f);
		conf >> f; mCraftSpecs.mModelOffset[YY] = Meter(f);
		conf >> f; mCraftSpecs.mModelOffset[ZZ] = Meter(f);
	}
	else if ( field == "ModelOffsetGear" )
	{
		fp f; conf >> f; mCraftSpecs.mModelOffsetGear = Meter(f);
	}
	else if ( field == "ModelRotationList" )
	{
		mCraftSpecs.mModelRotationList.Parse( conf );
	}
	else if ( field == "ModelLighting" )
	{
		string flag; conf >> flag;
		if ( flag == "false" or flag == "0" )
			mCraftSpecs.mModelLighting = false;
		else
			mCraftSpecs.mModelLighting = true;
	}

    // Errors?
	if ( error )
	{
		ostringstream msg;
		msg << "ERROR[conf]: parameter for " << field << " is invalid." << endl;
		throw Exception( msg.str() );
	}
}

} // namespace physics
