/*****************************************************************************
 * $LastChangedDate: 2008-06-01 00:41:19 -0400 (Sun, 01 Jun 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Physics model.
 *//*
 * LEGAL:	COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef PHYSICS_CRAFT_PHYSICS_HH
#define PHYSICS_CRAFT_PHYSICS_HH 1

#include "object/module.hh"
//#include "object/craft.hh"
namespace object { class Craft; }
using namespace object;
#include "physics/craft_specs.hh"

namespace physics {

////////////////////////////////////////////////////////////////////////////////
/// @brief CraftPhysics model.
///
/// Thrust/drag/lift:
/// The CraftPhysics class doesn't need to transform thrust/drag/lift (excluding weight).
/// These force vectors are relative to a Craft and therefore exist in local space.
/// Ultimately these are passed to Craft::Translate(Vector3) which transforms them.
///
/// Weight:
/// The direction of the weight vector is fixed/absolute of course.
/// Weight acceleration can be applied  using Object::GetPosition()/SetPosition()
/// to directly add weight vector in terms of a world vertex.
///
/// Compute() methods:
/// Many methods that just get a value are named Compute()
/// since ideally they would compute a physical parameter.
///
/// Craft state enums:
/// The state enums (flying, landing, etc) are kept in the Craft class.
/// Both Craft and CraftPhysics can affect the state.
///
class CraftPhysics
{
COPYABLE(CraftPhysics)

private:
	CLASS_CONST int THRUST_DIR = -1;  ///< multiplier/sign
	CLASS_CONST int LIFT_DIR   =  1;  ///< multiplier/sign

public:
                        CraftPhysics( SafePtr<Craft> craft, const CraftSpecs& specs );
                        ~CraftPhysics();

// Enable, reset:
// Physics can be globally disabled.
public:
	void                Enable( bool enable );
	bool				IfEnabled( void );
	void                Reset( void );

// Object animation:
public:
	void                Tick( const Milliseconds timeElapsed );
private:
    bool                AnimateCraft( const Milliseconds timeElapsed );
    bool                AnimateCraftLanding( const Milliseconds timeElapsed );

// Speed, velocity (Speed type is meters/second):
public:
	void				SetSpeed( const Speed speed );
    Speed	            ComputeSpeed( void );
	Speed				ComputeMaxSpeed( void );
	Velocity			ComputeVelocity( void ) { return mWorld.mVelocity; }
	Speed				ComputeStallSpeed( void ) { return KPH2Speed(mSpecs.mStallSpeed); }
	Speed				ComputeLiftoffSpeed( void ) { return KPH2Speed(mSpecs.mLiftoffSpeed); }
private:
	bool				IfVelocity( void );
	bool				IfForwardVelocity( const Matrix& craftMatrix );
	bool				IfDownwardVelocity( void );

// Thrust:
public:
	void                SetThrust( const Newton1 thrust );  // magnitude
	Newton1				ComputeMaxThrust( void ) { return mSpecs.mMaxThrustMag; }

// Drag:
public:
	void				EnableBrakes( const bool enable ) { mBrakes = enable; }
	bool				IfBrakes( void ) { return mBrakes; }
	bool				IfWheelBrakes( void );
	bool				IfAirBrakes( void ) { return IfBrakes() and not IfWheelBrakes(); }

// Fundamental forces:
public:
	Newton				ComputeThrustForce( void ) { return mWorld.mThrustForce; }
	Newton				ComputeDragForce( void )   { return mWorld.mDragForce; }
	Newton				ComputeLiftForce( void )   { return mWorld.mLiftForce; }
	Newton				ComputeWeightForce( void ) { return mWorld.mWeightForce; }

// Angle-of-attack, rotation rates:
public:
	Degree				ComputeAngleOfAttack( const bool ifVelocity );
	Degree				ComputeAngleOfAttack( void )
						{ return ComputeAngleOfAttack(IfVelocity()); }
	Degree				ComputeRollRate( void );
	Degree				ComputePitchRate( void );
	Degree				ComputeYawRate( void );
private:
	fp					ComputeRotationFactor( void );

// Turning:
public:
	enum eTurning { eNotTurning, eTurningLeft, eTurningRight };
	eTurning            ComputeTurningDir( void );
	bool				IfTurning( void );

// Stalling:
public:
	bool                IfStall( void );
	fp                  ComputeStall( void );

// Return a normalized vector of a Craft's matrix.
// Vector3 is produced from matrix elements in transposed order (vertical).
// Normalized in case matrix is scaled.
// Drag and weight are omitted since they aren't aligned with a craft.
private:
	Vector3				CraftAxisThrust( const Matrix& m )
						{ return Normalize<Vector3,fp>( m[Xz], m[Yz], m[Zz] ) * fp(THRUST_DIR); }
	Vector3				CraftAxisLift( const Matrix& m )
						{ return Normalize<Vector3,fp>( m[Xy], m[Yy], m[Zy] ) * fp(LIFT_DIR); }

private:
	////////////////////////////////////////////////////////////////////////////
	/// @brief Defines motion in a coordinate system.
	///
	/// MotionSpace is a vestige, but one that remains useful, from when
	/// PhysicsCraft used to do computations in both local and world spaces.
	///
	struct MotionSpace
	{
		MotionSpace( void );
		void Reset( void );

		Newton			mThrustForce, mDragForce, mLiftForce, mWeightForce;
		Acceleration	mThrustAccel, mDragAccel, mLiftAccel, mWeightAccel;
		Acceleration	mNetAccel;
		Velocity		mVelocity;
	};

/// Reference-counting problem:
/// CraftPhysics is a member of Craft.
/// A reference-counted class cannot have any reference to itself
/// else it would never be destroyed.  Therefore, CraftPhysics has
/// a SafePtr<Craft> back to Craft.
private:
	bool				mEnable;			///< enable this instance
    SafePtr<Craft>		mCraft;				///< (cannot be shptr!!) the Object this CraftPhysics instance controls
	CraftSpecs			mSpecs;				///< specifications/characteristics of a type of Craft
	Milliseconds		mPrevTick;
	MotionSpace			mWorld;				///< motion in world space
	Newton1				mThrustMag;			///< magnitude (current thrust)
	bool				mBrakes;
public:
	DECL_TYPESIG(TYPESIG_CRAFT_PHYSICS)
};

} // namespace physics

#endif // PHYSICS_CRAFT_PHYSICS_HH
