/*****************************************************************************
 * $LastChangedDate: 2008-09-12 08:28:20 -0400 (Fri, 12 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Aircraft classes.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PROGRAM_AIRCRAFT_CC 1
#include "base/module.hh"
using namespace base;
#include "glue/light_nums.hh"
using namespace glue;
#include "graph/module.hh"
#include "graph/subgraph.hh"
#include "graph/graph_visitors.hh"
#include "graph/light.hh"
using namespace graph;
#include "object/module.hh"
#include "object/dyna.hh"
#include "object/craft.hh"
using namespace object;
#include "physics/module.hh"
using namespace physics;
#include "control/module.hh"
#include "control/conf.hh"
using namespace control;
#include "fx/module.hh"
#include "fx/fx.hh"
using namespace fx;
#include "sound/module.hh"
using namespace sound;
#include "world/module.hh"
using namespace world;
#include "sky/module.hh"
using namespace sky;
#include "program/module.hh"
#include "program/aircraft.hh"
#include "program/carrier.hh"

namespace program {

INTERN const fp THROTTLE_FLAME = 0.7f;
INTERN const Milliseconds AIRCRAFT_EXPLOSION_LIFETIME( 5 * 60 * 1000 );  // 5 minutes

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  functions  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Make specified kind of aircraft.
 * Some 3D models were defined with its center at the nose.
 * physics_*.conf redefines an offset.
 * @param	aircraftType
 * @param	pos
 * @param	loadCopy
 *          Pass ModelCache::LOAD_SHARED or LOAD_COPY.
 *			To save memory, pass false for dummy (non-dynamic) aircraft
 *          such as those parked on the carrier.
 *****************************************************************************/
shptr<Aircraft>
MakeAircraft( Aircraft::eType aircraftType, const WorldVertex& pos, const bool loadCopy )
{
	shptr0<Aircraft> aircraft = NULL;
	shptr0<Graph>    graph    = NULL;

	// Aircraft need special lighting (esp. the SR-71 Blackbird).
	const WorldVertex lightPosition( GET_SKY().GetLight()->GetPosition() );
	uint			  lightNum( glue::defs::LIGHT_NUM_MODEL );
	RGBA			  lightAmbient(  0.15f, 0.15f, 0.15f, 1.00f );
	RGBA			  lightDiffuse(  0.15f, 0.15f, 0.15f, 1.00f );
	RGBA			  lightSpecular( 0.70f, 0.70f, 0.70f, 1.00f );
	bool			  baseAmbient = true;

#define MAKE_AIRCRAFT( CLASS )													\
{																				\
	graph = program::LoadModel( CLASS::ClassGetCraftSpecs(), loadCopy ).PTR();	\
	aircraft = new CLASS( graph.PTR(), pos );									\
}

	// For a type of Aircraft:
	switch ( aircraftType )
	{
		case Aircraft::eType_Spitfire:
		{
			MAKE_AIRCRAFT( SpitfireAircraft );
		}
		break;

		case Aircraft::eType_Corsair:
		{
			MAKE_AIRCRAFT( CorsairAircraft );
		}
		break;

		case Aircraft::eType_P51:
		{
			MAKE_AIRCRAFT( P51Aircraft );
		}
		break;

		case Aircraft::eType_F82:
		{
			MAKE_AIRCRAFT( F82Aircraft );
		}
		break;

		case Aircraft::eType_F86:
		{
			MAKE_AIRCRAFT( F86Aircraft );
		}
		break;

		case Aircraft::eType_A4:
		{
			MAKE_AIRCRAFT( A4Aircraft );
		}
		break;

		case Aircraft::eType_F15:
		{
			MAKE_AIRCRAFT( F15Aircraft );
		}
		break;

		case Aircraft::eType_F16:
		{
			MAKE_AIRCRAFT( F16Aircraft );
		}
		break;

		case Aircraft::eType_F18:
		{
			MAKE_AIRCRAFT( F18Aircraft );
		}
		break;

		case Aircraft::eType_SR71:
		{
			// Very dark and high specular for Blackbird.
			lightNum = glue::defs::LIGHT_NUM_MODEL_SR71;
			lightAmbient  = RGBA( 0.02f, 0.02f, 0.02f, 1.00f );
			lightDiffuse  = RGBA( 0.02f, 0.02f, 0.02f, 1.00f );
			lightSpecular = RGBA( 0.40f, 0.40f, 0.40f, 1.00f );
			baseAmbient = false;
			MAKE_AIRCRAFT( SR71Aircraft );
		}
		break;

		case Aircraft::eType_Mirage2000:
		{
			MAKE_AIRCRAFT( Mirage2000Aircraft );
		}
		break;

		case Aircraft::eType_SU37:
		{
			MAKE_AIRCRAFT( SU37Aircraft );
		}
		break;

		case Aircraft::eType_Sikorsky:
		{
			MAKE_AIRCRAFT( SikorskyAircraft );
		}
		break;

		default:
		{
			ASSERT(0);  // oops
		}
		break;
	}

	// Commit light source.
	shptr<Light> light = \
		program::InsertLightIntoGraph( graph.PTR(),
                                       lightNum,
                                       lightAmbient,
                                       lightDiffuse,
                                       lightSpecular );
	if ( not baseAmbient )  // for SR-71
		light->SetBaseAmbient( RGBA( 0.0f, 0.0f, 0.0f, 0.0f ) );

	// Return Aircraft.
	return aircraft.PTR();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  Aircraft  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Aircraft::Aircraft( shptr<Graph> graph, const WorldVertex& pos,
                    const CraftSpecs& craftSpecs )
:	Parent(graph,pos,craftSpecs)
{
	// The rotation list orients 3D models how the simulato expects (pitch,yaw,roll).
	SET_TYPESIG(this,TYPESIG_AIRCRAFT);
	PROGRAM_CONF.mAircraftRotationList.Apply( *this );
	SetThrottle( 0.0f );
}

Aircraft::~Aircraft()
{
	INVALIDATE_TYPESIG(this,TYPESIG_AIRCRAFT);
}

/*****************************************************************************
 * Reset.
 *****************************************************************************/
void
Aircraft::Reset( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Reset();

	EnableBrakes( false );

	// Clear collision.
	mCollision = false;

	// Let it fly again.
	GetCraftPhysics().Enable( true );
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
Aircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	MoreCrashDetermination();
}

/*****************************************************************************
 * The physics model determines one cause of a crash (crash-dive),
 * but other causes are possible.
 *****************************************************************************/
void
Aircraft::MoreCrashDetermination( void )
{
	// Determine if a collision, detected earlier, is a craft.
	if ( IfCollision() and not IfCrash() )
	{
		// What Object did this Aircraft collide into?
		shptr0<Object> collisionObject = GetCollisionObject();
		ASSERT( collisionObject != NULL );

		// CraftPhysics catches if Aircraft collides too fast into carrier.
		// At this point, Aircraft has either "safely collided" into carrier
		// or collided into another Object.
		if ( collisionObject->GetName() == "NimitzCarrier" )
		{
			// Aircraft collided into carrier.
			// Colliding into runway is safe.
			// But did Aircraft collide into another part of the carrier?
			SafePtr<Carrier> carrier = reinterpret_cast<Carrier*>( collisionObject.PTR() );
			CHECK_TYPESIG(carrier,TYPESIG_CARRIER);
			if ( not carrier->IfRunwayAltitude(GetPosition()) )
			{
				// Crashed below flight deck or into the tower.
				SetState( eState_CRASHED );
				SetCauseOfCrash( eCauseOfCrash_DEFAULT );
			}
		}
		else
		{
			// Collided into another Object (excluding carrier).
			SetState( eState_CRASHED );
			SetCauseOfCrash( eCauseOfCrash_DEFAULT );
		}
	}
}

/*****************************************************************************
 * Set state.
 *****************************************************************************/
void
Aircraft::SetState( const eState state )
{
	// If transition to CRASHED, make an explosion once.
	if ( (not IfCrash()) and (state == eState_CRASHED) )
	{
		// Visual explosion.
		GET_FX().MakeExplosion( GetPosition(), GetRadius(), AIRCRAFT_EXPLOSION_LIFETIME );

		// Crash sound.
		if ( this == GET_CURRENT_CRAFT().PTR() )
		{
			StopEngineSound();
			sound::Sound::GetInstance().Play( "crash.wav.gz", sound::Sound::PLAY_ONCE );
		}

		// Stop engine/propeller.
		SetThrottle( 0.0f );
	}

	// Afterwards since this affects IfCrash() used above.
	Parent::SetState( state );
}

/*****************************************************************************
 * @return Height in world space of Object.
 *****************************************************************************/
fp
Aircraft::GetHeight( void )
{
	return Geo::Meters2Sim( GetCraftSpecs().mHeight );
}

/*****************************************************************************
 * Play engine sound.  Controlled by Lua.
 *****************************************************************************/
void
Aircraft::PlayEngineSound( const fp volume )
{
	if ( this == GET_CURRENT_CRAFT().PTR() )
	{
		// Play sound for piston or jet engine.
		if ( IfHasPropeller() )
		{
			// If no engine sound yet, play engine-cranking sound,
			// unless playing begin flying.
			if ( not sound::Sound::GetInstance().IfPlaying( "piston.wav.gz" )
	         and Speed2KPH(GetCraftPhysics().ComputeSpeed()) < SpeedKPH(2) )
			{
				sound::Sound::GetInstance().Play( "piston_cough.wav.gz",
			                                      sound::Sound::PLAY_ONCE,
				                                  1.0f );
			}
	
			sound::Sound::GetInstance().Play( "piston.wav.gz",
	                                          sound::Sound::PLAY_LOOP,
	                                          volume );
		}
		else
		{
			sound::Sound::GetInstance().Play( "jet.wav.gz",
	                                          sound::Sound::PLAY_LOOP,
	                                          volume );
		}
	}
}

/*****************************************************************************
 * Stop engine sound.
 *****************************************************************************/
void
Aircraft::StopEngineSound( void )
{
	// These loop.
	if ( this == GET_CURRENT_CRAFT().PTR() )
	{
		sound::Sound::GetInstance().Stop( "piston.wav.gz" );
		sound::Sound::GetInstance().Stop( "jet.wav.gz" );
	}
}

/*****************************************************************************
 * @return Try if Aircraft is flying.
 *****************************************************************************/
bool
Aircraft::IfFlying( void )
{
	return Speed2KPH(GetCraftPhysics().ComputeSpeed()) > SpeedKPH(10);
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// IfCanSetLandingGear() was already called by derivative.

	Parent::SetLandingGear( down );

	// No sound when initially setting landing gear on runway.
	if ( this == GET_CURRENT_CRAFT().PTR() and IfFlying() )
	{
		sound::Sound::GetInstance().Play( "gear.wav.gz", Sound::PLAY_ONCE );
	}
}

/*****************************************************************************
 * Prevent silliness such as retracting landing gear on runway.
 *****************************************************************************/
bool
Aircraft::IfCanSetLandingGear( const bool down )
{
	return down or IfFlying();
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  SpitfireAircraft  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SpitfireAircraft::SpitfireAircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,SpitfireAircraft::ClassGetCraftSpecs()),
	mSubgraphPropeller1(new Subgraph(graph,"Propeller1")),
	mSubgraphPropeller2(new Subgraph(graph,"Propeller2")),
	mSubgraphPropeller3(new Subgraph(graph,"Propeller3")),
	mSubgraphPropeller4(new Subgraph(graph,"Propeller4")),
	mSubgraphGear0(new Subgraph(graph,"Leg-Assembly-L")),
	mSubgraphGear1(new Subgraph(graph,"Leg-Assembly-R")),
	mSubgraphGear2(new Subgraph(graph,"Tail-Wheel-Assembly"))
{
	// NOP
}

SpitfireAircraft::~SpitfireAircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
SpitfireAircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	return SpitfireAircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
SpitfireAircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_spitfire.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
SpitfireAircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	RotatePropeller();
}

/*****************************************************************************
 * Rotate propeller.
 *****************************************************************************/
// Rotate both propeller Groups.
void
SpitfireAircraft::RotatePropeller( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// Degree of rotation correlates to throttle.
	const Degree degree = 35.0f * GetThrottle();
	const Radian radian = Deg2Rad( -degree );
	mSubgraphPropeller1->Rotate( XX, radian );
	mSubgraphPropeller2->Rotate( XX, radian );
	mSubgraphPropeller3->Rotate( XX, radian );
	mSubgraphPropeller4->Rotate( XX, radian );
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
SpitfireAircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
		mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  CorsairAircraft  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
CorsairAircraft::CorsairAircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,CorsairAircraft::ClassGetCraftSpecs()),
	mSubgraphPropeller0(new Subgraph(graph,"prop_0")),
	mSubgraphPropeller1(new Subgraph(graph,"prop_1")),
	mSubgraphGear0(new Subgraph(graph,"gearp2.L")),
	mSubgraphGear1(new Subgraph(graph,"gearp2.R")),
	mSubgraphGear2(new Subgraph(graph,"gearp3.L")),
	mSubgraphGear3(new Subgraph(graph,"gearp3.R")),
	mSubgraphGear4(new Subgraph(graph,"gearcylinder.L")),
	mSubgraphGear5(new Subgraph(graph,"gearcylinder.R")),
	mSubgraphGear6(new Subgraph(graph,"gearleg.L")),
	mSubgraphGear7(new Subgraph(graph,"gearleg.R")),
	mSubgraphGear8(new Subgraph(graph,"gearleg2.L")),
	mSubgraphGear9(new Subgraph(graph,"gearleg2.R")),
	mSubgraphGear10(new Subgraph(graph,"gearp1.L")),
	mSubgraphGear11(new Subgraph(graph,"gearp1.R")),
	mSubgraphGear12(new Subgraph(graph,"gearsc1.L")),
	mSubgraphGear13(new Subgraph(graph,"gearsc2.L")),
	mSubgraphGear14(new Subgraph(graph,"tailgear1")),
	mSubgraphGear15(new Subgraph(graph,"tailgear2")),
	mSubgraphGear16(new Subgraph(graph,"stopwheel.L")),
	mSubgraphGear17(new Subgraph(graph,"stopwheel.R")),
	mSubgraphGear18(new Subgraph(graph,"wheel.L")),
	mSubgraphGear19(new Subgraph(graph,"wheel.R")),
	mSubgraphGear20(new Subgraph(graph,"tailwheel"))
{
	// NOP
}

CorsairAircraft::~CorsairAircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
CorsairAircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return CorsairAircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
CorsairAircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_corsair.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
CorsairAircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	RotatePropeller();
}

/*****************************************************************************
 * Rotate propeller.
 *****************************************************************************/
void
CorsairAircraft::RotatePropeller( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// Degree of rotation correlates to throttle.
	const Degree degree = 35.0f * GetThrottle();
	const Radian radian = Deg2Rad( -degree );
	mSubgraphPropeller0->Rotate( XX, radian );
	mSubgraphPropeller1->Rotate( XX, radian );
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
CorsairAircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down ); mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down ); mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down ); mSubgraphGear5->Enable( down );
		mSubgraphGear6->Enable( down ); mSubgraphGear7->Enable( down );
		mSubgraphGear8->Enable( down ); mSubgraphGear9->Enable( down );
		mSubgraphGear10->Enable( down ); mSubgraphGear11->Enable( down );
		mSubgraphGear12->Enable( down ); mSubgraphGear13->Enable( down );
		mSubgraphGear14->Enable( down ); mSubgraphGear15->Enable( down );
		mSubgraphGear16->Enable( down ); mSubgraphGear17->Enable( down );
		mSubgraphGear18->Enable( down ); mSubgraphGear19->Enable( down );
		mSubgraphGear20->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  P51Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
P51Aircraft::P51Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,P51Aircraft::ClassGetCraftSpecs()),
	mSubgraphPropeller1(new Subgraph(graph,"Propeller1")),
	mSubgraphPropeller2(new Subgraph(graph,"Propeller2")),
	//
	mSubgraphLeftGearWheel(new Subgraph(graph,"LeftGearWheel")),
	mSubgraphLeftGearStrut(new Subgraph(graph,"LeftGearStrut")),
	mSubgraphLeftInnerGearCover(new Subgraph(graph,"LeftInnerGearCover")),
	mSubgraphLeftOuterGearCover(new Subgraph(graph,"LeftOuterGearCover")),
	mSubgraphLeftTailGearDoor(new Subgraph(graph,"LeftTailGearDoor")),
	//
	mSubgraphRightGearWheel(new Subgraph(graph,"RightGearWheel")),
	mSubgraphRightGearStrut(new Subgraph(graph,"RightGearStrut")),
	mSubgraphRightInnerGearCover(new Subgraph(graph,"RightInnerGearCover")),
	mSubgraphRightOuterGearCover(new Subgraph(graph,"RightOuterGearCover")),
	mSubgraphRightTailGearDoor(new Subgraph(graph,"RightTailGearDoor")),
	//
	mSubgraphTailGearWheel(new Subgraph(graph,"TailGearWheel")),
	mSubgraphTailGearStrut(new Subgraph(graph,"TailGearStrut"))
{
	// NOP
}

P51Aircraft::~P51Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
P51Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return P51Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
P51Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_p51.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
P51Aircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	RotatePropeller();
}

/*****************************************************************************
 * Rotate propeller.
 *****************************************************************************/
void
P51Aircraft::RotatePropeller( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// Rotate both propeller Groups.
	// Degree of rotation correlates to throttle.
	const Degree degree = 35.0f * GetThrottle();
	const Radian radian = Deg2Rad( -degree );
	mSubgraphPropeller1->Rotate( XX, radian );  // this p51d-jw model splits propeller
	mSubgraphPropeller2->Rotate( XX, radian );  // into two subgraphs
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
P51Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// The p51d-nw model apparently defines the origin of the landing gear
	// at the center of the fuselage instead of the strut pivot.
	// For now, just disable the nodes instead of doing rotation.

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphLeftGearWheel->Enable( down );
		mSubgraphLeftGearStrut->Enable( down );
		mSubgraphLeftInnerGearCover->Enable( down );
		mSubgraphLeftOuterGearCover->Enable( down );
		mSubgraphLeftTailGearDoor->Enable( down );

		mSubgraphRightGearWheel->Enable( down );
		mSubgraphRightGearStrut->Enable( down );
		mSubgraphRightInnerGearCover->Enable( down );
		mSubgraphRightOuterGearCover->Enable( down );
		mSubgraphRightTailGearDoor->Enable( down );

		mSubgraphTailGearWheel->Enable( down );
		mSubgraphTailGearStrut->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  F82Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
F82Aircraft::F82Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,F82Aircraft::ClassGetCraftSpecs()),
	mSubgraphPropeller1(new Subgraph(graph,"Propeller1")),
	mSubgraphPropeller2(new Subgraph(graph,"Propeller2")),
	mSubgraphPropeller3(new Subgraph(graph,"Propeller1.002")),
	mSubgraphPropeller4(new Subgraph(graph,"Propeller2.001")),
	//
	mSubgraphLeftGearWheel(new Subgraph(graph,"LeftGearWheel")),
	mSubgraphLeftGearStrut(new Subgraph(graph,"LeftGearStrut")),
	mSubgraphLeftInnerGearCover(new Subgraph(graph,"LeftInnerGearCover")),
	mSubgraphLeftInnerGearCover2(new Subgraph(graph,"LeftInnerGearCove")),  // sic/truncation
	mSubgraphLeftOuterGearCover(new Subgraph(graph,"LeftOuterGearCover")),
	mSubgraphLeftTailGearDoor(new Subgraph(graph,"LeftTailGearDoor")),
	//
	mSubgraphRightGearWheel(new Subgraph(graph,"RightGearWheel")),
	mSubgraphRightGearStrut(new Subgraph(graph,"RightGearStrut")),
	mSubgraphRightInnerGearCover(new Subgraph(graph,"RightInnerGearCover")),
	mSubgraphRightInnerGearCover2(new Subgraph(graph,"RightInnerGearCov")),  // sic/truncation
	mSubgraphRightOuterGearCover(new Subgraph(graph,"RightOuterGearCover")),
	mSubgraphRightTailGearDoor(new Subgraph(graph,"RightTailGearDoor")),
	//
	mSubgraphTailGearWheel(new Subgraph(graph,"TailGearWheel")),
	mSubgraphTailGearStrut(new Subgraph(graph,"TailGearStrut")),
	mSubgraphTailGearWheel2(new Subgraph(graph,"TailGearWheel.001")),
	mSubgraphTailGearStrut2(new Subgraph(graph,"TailGearStrut.001"))
{
	// NOP
}

F82Aircraft::~F82Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
F82Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return F82Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
F82Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_f82.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
F82Aircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	RotatePropeller();
}

/*****************************************************************************
 * Rotate propeller.
 *****************************************************************************/
void
F82Aircraft::RotatePropeller( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// Rotate both propeller Groups.
	// Degree of rotation correlates to throttle.
	const Degree degree = 35.0f * GetThrottle();
	const Radian radian = Deg2Rad( -degree );

	// The origin of the propeller's subgraph isn't positioned correctly for rotation,
	// so just enable/disable instead.  The 3D model needs to be edited.
	const bool enable = GetThrottle() < 0.01f;
	mSubgraphPropeller1->Enable( enable );
	mSubgraphPropeller2->Enable( enable );
	mSubgraphPropeller3->Enable( enable );
	mSubgraphPropeller4->Enable( enable );
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
F82Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// The model apparently defines the origin of the landing gear
	// at the center of the fuselage instead of the strut pivot.
	// For now, just disable the nodes instead of doing rotation.

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphLeftGearWheel->Enable( down );
		mSubgraphLeftGearStrut->Enable( down );
		mSubgraphLeftInnerGearCover->Enable( down );
		mSubgraphLeftInnerGearCover2->Enable( down );
		mSubgraphLeftOuterGearCover->Enable( down );
		mSubgraphLeftTailGearDoor->Enable( down );

		mSubgraphRightGearWheel->Enable( down );
		mSubgraphRightGearStrut->Enable( down );
		mSubgraphRightInnerGearCover->Enable( down );
		mSubgraphRightInnerGearCover2->Enable( down );
		mSubgraphRightOuterGearCover->Enable( down );
		mSubgraphRightTailGearDoor->Enable( down );

		mSubgraphTailGearWheel->Enable( down );
		mSubgraphTailGearStrut->Enable( down );
		mSubgraphTailGearWheel2->Enable( down );
		mSubgraphTailGearStrut2->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  F86Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
F86Aircraft::F86Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,F86Aircraft::ClassGetCraftSpecs()),
	mSubgraphGear0(new Subgraph(graph,"nosegear1")),
	mSubgraphGear1(new Subgraph(graph,"nosewheel")),
	mSubgraphGear2(new Subgraph(graph,"gear.L")),
	mSubgraphGear3(new Subgraph(graph,"gear.R")),
	mSubgraphGear4(new Subgraph(graph,"wheel.L")),
	mSubgraphGear5(new Subgraph(graph,"wheel.R"))
	// (no flame)
{
	SetThrottle( 0.0f );
}

F86Aircraft::~F86Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
F86Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return F86Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
F86Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_f86.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
F86Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );  mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );  mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );  mSubgraphGear5->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  A4Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
A4Aircraft::A4Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,A4Aircraft::ClassGetCraftSpecs()),
	mSubgraphGear0(new Subgraph(graph,"LeftWheel")),
	mSubgraphGear1(new Subgraph(graph,"LeftLowerLeg")),
	mSubgraphGear2(new Subgraph(graph,"LeftUpperLeg")),
	mSubgraphGear3(new Subgraph(graph,"LeftGearDoor")),
	mSubgraphGear4(new Subgraph(graph,"LeftGearStrut")),
	mSubgraphGear5(new Subgraph(graph,"LeftLowerLegAssmbly")),
	mSubgraphGear6(new Subgraph(graph,"RightWheel")),
	mSubgraphGear7(new Subgraph(graph,"RightLowerLeg")),
	mSubgraphGear8(new Subgraph(graph,"RightUpperLeg")),
	mSubgraphGear9(new Subgraph(graph,"RightGearDoor")),
	mSubgraphGear10(new Subgraph(graph,"RightGearStrut")),
	mSubgraphGear11(new Subgraph(graph,"RightLowerLegAssmbly")),
	mSubgraphGear12(new Subgraph(graph,"LandingLight")),    // unique to left wheel
	mSubgraphGear13(new Subgraph(graph,"NoseWheel")),       // front wheel
	mSubgraphGear14(new Subgraph(graph,"Upper-Leg-Nose")),  // front wheel
	mSubgraphGear15(new Subgraph(graph,"Lower-Leg")),       // front wheel
	mSubgraphGear16(new Subgraph(graph,"NoseGearStrut")),   // front wheel
	mSubgraphGear17(new Subgraph(graph,"NoseGearDoor"))     // front wheel
// (no flame)
{
	SetThrottle( 0.0f );
}

A4Aircraft::~A4Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
A4Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return A4Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
A4Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_a4.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
A4Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );  mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );  mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );  mSubgraphGear5->Enable( down );
		mSubgraphGear6->Enable( down );  mSubgraphGear7->Enable( down );
		mSubgraphGear8->Enable( down );  mSubgraphGear9->Enable( down );
		mSubgraphGear10->Enable( down );  mSubgraphGear11->Enable( down );
		mSubgraphGear12->Enable( down );  mSubgraphGear13->Enable( down );
		mSubgraphGear14->Enable( down );  mSubgraphGear15->Enable( down );
		mSubgraphGear16->Enable( down );  mSubgraphGear17->Enable( down );
	}
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  F15Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
F15Aircraft::F15Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,F15Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"LeftWheel")),
	mSubgraphGear1(new Subgraph(graph,"LeftWheelTire")),
	mSubgraphGear2(new Subgraph(graph,"RightWheel")),
	mSubgraphGear3(new Subgraph(graph,"RightWheelTire")),
	mSubgraphGear4(new Subgraph(graph,"FrontWheel")),
	mSubgraphGear5(new Subgraph(graph,"FrontWheelTire")),
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"InternalFlame")),
	mSubgraphFlame1(new Subgraph(graph,"ExternalFlame"))
{
	SetThrottle( 0.0f );
}

F15Aircraft::~F15Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
F15Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return F15Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
F15Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_f15.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
F15Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
		mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );
		mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );
		mSubgraphGear5->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
F15Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
	mSubgraphFlame0->Enable( flame );
	mSubgraphFlame1->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  F16Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
F16Aircraft::F16Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,F16Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"internal_front_gear_door")),
	mSubgraphGear1(new Subgraph(graph,"internal_right_main_door")),
	mSubgraphGear2(new Subgraph(graph,"internal_left_main_door")),
	//
	mSubgraphGear3(new Subgraph(graph,"front_wheel")),
	mSubgraphGear4(new Subgraph(graph,"front_aft_strut")),
	mSubgraphGear5(new Subgraph(graph,"front_lower_strut")),
	mSubgraphGear6(new Subgraph(graph,"front_upper_strut")),
	//
	mSubgraphGear7(new Subgraph(graph,"left_main_strut")),
	mSubgraphGear8(new Subgraph(graph,"left_main_wheel")),
	mSubgraphGear9(new Subgraph(graph,"left_lower_main_strut")),
	mSubgraphGear10(new Subgraph(graph,"left_upper_main_strut")),
	//
	mSubgraphGear11(new Subgraph(graph,"right_main_strut")),
	mSubgraphGear12(new Subgraph(graph,"right_main_wheel")),
	mSubgraphGear13(new Subgraph(graph,"right_lower_main_strut")),
	mSubgraphGear14(new Subgraph(graph,"right_upper_main_strut")),
	//
	mSubgraphGear15(new Subgraph(graph,"external_front_gear_door")),
	mSubgraphGear16(new Subgraph(graph,"external_left_main_door")),
	mSubgraphGear17(new Subgraph(graph,"external_right_main_door")),
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"internal_flame")),
	mSubgraphFlame1(new Subgraph(graph,"external_flame")),
	mSubgraphFlame2(new Subgraph(graph,"afterburner"))
{
	SetThrottle( 0.0f );
}

F16Aircraft::~F16Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
F16Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return F16Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
F16Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_f16.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
F16Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	// Known problem: The door for the front wheel should be rotated
	// into the fuselage but isn't implemented.

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );  mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );  mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );  mSubgraphGear5->Enable( down );
		mSubgraphGear6->Enable( down );  mSubgraphGear7->Enable( down );
		mSubgraphGear8->Enable( down );  mSubgraphGear9->Enable( down );
		mSubgraphGear10->Enable( down ); mSubgraphGear11->Enable( down );
		mSubgraphGear12->Enable( down ); mSubgraphGear13->Enable( down );
		mSubgraphGear14->Enable( down ); mSubgraphGear15->Enable( down );
	  //mSubgraphGear16->Enable( down ); mSubgraphGear17->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
F16Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
//	mSubgraphFlame0->Enable( flame );  // to cover somewhat hold in model
	mSubgraphFlame1->Enable( flame );
	mSubgraphFlame2->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  F18Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
F18Aircraft::F18Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,F18Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"LeftWheel")),
	mSubgraphGear1(new Subgraph(graph,"LeftDoor")),
	mSubgraphGear2(new Subgraph(graph,"RightWheel")),
	mSubgraphGear3(new Subgraph(graph,"RightDoor")),
	mSubgraphGear4(new Subgraph(graph,"FrontWheel")),
	mSubgraphGear5(new Subgraph(graph,"FrontDoor")),
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"InternalFlame")),
	mSubgraphFlame1(new Subgraph(graph,"ExternalFlame"))
{
	SetThrottle( 0.0f );
}

F18Aircraft::~F18Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
F18Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return F18Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
F18Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_f18.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
F18Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) ) 
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
		mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );
		mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );
		mSubgraphGear5->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
F18Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
	mSubgraphFlame0->Enable( flame );
	mSubgraphFlame1->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  SR71Aircraft  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SR71Aircraft::SR71Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,SR71Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"roueA")),  // roue means wheel in French
	mSubgraphGear1(new Subgraph(graph,"roueD")),
	mSubgraphGear2(new Subgraph(graph,"roueG")),
	// struts:
	mSubgraphStrut0(new Subgraph(graph,"axeA")),
	mSubgraphStrut1(new Subgraph(graph,"axeD")),
	mSubgraphStrut2(new Subgraph(graph,"axeG")),
	mSubgraphStrut3(new Subgraph(graph,"barrebasA")),
	mSubgraphStrut4(new Subgraph(graph,"barrebasD")),
	mSubgraphStrut5(new Subgraph(graph,"barrebasG")),
	mSubgraphStrut6(new Subgraph(graph,"barrehautA")),
	mSubgraphStrut7(new Subgraph(graph,"barrehautD")),
	mSubgraphStrut8(new Subgraph(graph,"barrehautG")),
	mSubgraphStrut9(new Subgraph(graph,"porteD1")),
	mSubgraphStrut10(new Subgraph(graph,"porteD2")),
	mSubgraphStrut11(new Subgraph(graph,"porteG1")),
	mSubgraphStrut12(new Subgraph(graph,"porteG2")),
	mSubgraphStrut13(new Subgraph(graph,"verrinA")),
	mSubgraphStrut14(new Subgraph(graph,"verrinD")),
	mSubgraphStrut15(new Subgraph(graph,"verrinG")),
	// parachute:
	mSubgraphParachute(new Subgraph(graph,"parachute")),
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"flammeD1")),
	mSubgraphFlame1(new Subgraph(graph,"flammeD2")),
	mSubgraphFlame2(new Subgraph(graph,"flammeD3")),
	mSubgraphFlame3(new Subgraph(graph,"flammeG1")),
	mSubgraphFlame4(new Subgraph(graph,"flammeG2")),
	mSubgraphFlame5(new Subgraph(graph,"flammeG3"))
{
	SetThrottle( 0.0f );
	mSubgraphParachute->Enable( false );
}

SR71Aircraft::~SR71Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
SR71Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return SR71Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
SR71Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_sr71.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
SR71Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down )) 
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
		mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );

		mSubgraphStrut0->Enable( down );
		mSubgraphStrut1->Enable( down );
		mSubgraphStrut2->Enable( down );
		mSubgraphStrut3->Enable( down );
		mSubgraphStrut4->Enable( down );
		mSubgraphStrut5->Enable( down );
		mSubgraphStrut6->Enable( down );
		mSubgraphStrut7->Enable( down );
		mSubgraphStrut8->Enable( down );
		mSubgraphStrut9->Enable( down );
		mSubgraphStrut10->Enable( down );
		mSubgraphStrut11->Enable( down );
		mSubgraphStrut12->Enable( down );
		mSubgraphStrut13->Enable( down );
		mSubgraphStrut14->Enable( down );
		mSubgraphStrut15->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
SR71Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
	mSubgraphFlame0->Enable( flame );
	mSubgraphFlame1->Enable( flame );
	mSubgraphFlame2->Enable( flame );
	mSubgraphFlame3->Enable( flame );
	mSubgraphFlame4->Enable( flame );
	mSubgraphFlame5->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  Mirage2000Aircraft  //////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
Mirage2000Aircraft::Mirage2000Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,Mirage2000Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"front_tire")),
	mSubgraphGear1(new Subgraph(graph,"front_wheel")),
	mSubgraphGear2(new Subgraph(graph,"front_door_fwd")),
	mSubgraphGear3(new Subgraph(graph,"left_wheel")),
	mSubgraphGear4(new Subgraph(graph,"left_tire")),
	mSubgraphGear5(new Subgraph(graph,"left_door")),
	mSubgraphGear6(new Subgraph(graph,"right_wheel")),
	mSubgraphGear7(new Subgraph(graph,"right_tire")),
	mSubgraphGear8(new Subgraph(graph,"right_door")),
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"internal_flame")),
	mSubgraphFlame1(new Subgraph(graph,"external_flame"))
{
	SetThrottle( 0.0f );
}

Mirage2000Aircraft::~Mirage2000Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
Mirage2000Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return Mirage2000Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
Mirage2000Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_mirage2000.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
Mirage2000Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down )) 
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
		mSubgraphGear1->Enable( down );
		mSubgraphGear2->Enable( down );
		mSubgraphGear3->Enable( down );
		mSubgraphGear4->Enable( down );
		mSubgraphGear5->Enable( down );
		mSubgraphGear6->Enable( down );
		mSubgraphGear7->Enable( down );
		mSubgraphGear8->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
Mirage2000Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
	mSubgraphFlame0->Enable( flame );
	mSubgraphFlame1->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////////  SU37Aircraft  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SU37Aircraft::SU37Aircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,SU37Aircraft::ClassGetCraftSpecs()),
	// gear:
	mSubgraphGear0(new Subgraph(graph,"UC")),  // under carriage
	// exhaust flame:
	mSubgraphFlame0(new Subgraph(graph,"L_Reheat")),
	mSubgraphFlame1(new Subgraph(graph,"R_Reheat"))
{
	SetThrottle( 0.0f );
}

SU37Aircraft::~SU37Aircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
SU37Aircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);
	return SU37Aircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
SU37Aircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_su37.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
SU37Aircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down )) 
	{
		Parent::SetLandingGear( down );

		mSubgraphGear0->Enable( down );
	}
}

/*****************************************************************************
 * When throttle is changed, animate jet exhaust flame.
 *****************************************************************************/
void
SU37Aircraft::SetThrottle( const fp throttle )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::SetThrottle( throttle );
	const bool flame = (throttle >= THROTTLE_FLAME);
	mSubgraphFlame0->Enable( flame );
	mSubgraphFlame1->Enable( flame );
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////  SikorskyAircraft  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
SikorskyAircraft::SikorskyAircraft( shptr<Graph> graph, const WorldVertex& pos )
:	Parent(graph,pos,SikorskyAircraft::ClassGetCraftSpecs()),
	mSubgraphRotor0(new Subgraph(graph,"blade1")),
	mSubgraphRotor1(new Subgraph(graph,"blade2")),
	mSubgraphRotor2(new Subgraph(graph,"blade3")),
	mSubgraphRotor3(new Subgraph(graph,"blade4")),
	mSubgraphRotor4(new Subgraph(graph,"fastblade1")),  // blurred
	mSubgraphRotor5(new Subgraph(graph,"fastblade2")),  // blurred
	mSubgraphRotor6(new Subgraph(graph,"fastblade3")),  // blurred
	mSubgraphRotor7(new Subgraph(graph,"fastblade4")),  // blurred
	mSubgraphTailRotor(new Subgraph(graph,"TailRotor"))
{
	// NOP
}

SikorskyAircraft::~SikorskyAircraft()
{
	// NOP
}

/*****************************************************************************
 * @return Aircraft specifications struct.
 *****************************************************************************/
const CraftSpecs&
SikorskyAircraft::GetCraftSpecs( void )  // virtual method
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	return SikorskyAircraft::ClassGetCraftSpecs();
}

const CraftSpecs&
SikorskyAircraft::ClassGetCraftSpecs( void )  // class method
{
	// Load specs from a .conf file once.
	PERSISTENT CraftSpecs sCraftSpecs;
	BEGIN_ONCE {
		sCraftSpecs.ReadSpecsFromConfFile( "physics_sikorsky.conf" );
	} END_ONCE
	return sCraftSpecs;
}

/*****************************************************************************
 * Tick that pulses animation.
 *****************************************************************************/
void
SikorskyAircraft::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	Parent::Tick( millisecElapsed );

	RotateRotors();
}

/*****************************************************************************
 * Rotate propeller.
 *****************************************************************************/
void
SikorskyAircraft::RotateRotors( void )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	const uint   axisRotor     = YY;  // main rotors
//	const uint   axisTailRotor = ZZ;
  {
	const Radian radian = Deg2Rad( Degree(10.0f) );
	mSubgraphRotor0->Rotate( axisRotor, radian );
	mSubgraphRotor1->Rotate( axisRotor, radian );
	mSubgraphRotor2->Rotate( axisRotor, radian );
	mSubgraphRotor3->Rotate( axisRotor, radian );
  }
  {
	const Radian radian = Deg2Rad( Degree(5.0f) );  // blurred rotors
	mSubgraphRotor4->Rotate( axisRotor, radian );
	mSubgraphRotor5->Rotate( axisRotor, radian );
	mSubgraphRotor6->Rotate( axisRotor, radian );
	mSubgraphRotor7->Rotate( axisRotor, radian );
  }

	// The origin of the tail rotor is the center of the helicopter
	// rather than the rotor's hub, so just toggle enablement.
#if 1
	mSubgraphTailRotor->Enable( not mSubgraphTailRotor->IfEnabled() );
#else
  {
	const Radian radian = Deg2Rad( Degree(30.0f) );
	mSubgraphTailRotor->Rotate( axisTailRotor, radian );
  }
#endif
}

/*****************************************************************************
 * Retract/extend landing gear.
 *****************************************************************************/
void
SikorskyAircraft::SetLandingGear( const bool down )
{
CHECK_TYPESIG(this,TYPESIG_AIRCRAFT);

	if ( IfCanSetLandingGear( down ) )
	{
		Parent::SetLandingGear( down );

		Parent::mLandingGearDown = true;  // override
	}
}

} // namespace program
