/*****************************************************************************
 * $LastChangedDate: 2008-08-31 02:29:07 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Aircraft classes.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef PROGRAM_AIRCRAFT_HH
#define PROGRAM_AIRCRAFT_HH 1

#include "graph/module.hh"
#include "graph/subgraph.hh"
using namespace graph;

namespace program {

////////////////////////////////////////////////////////////////////////////////
/// @brief Aircraft.
///
/// This class is unusual:
/// - Aircraft has both a virtual and class method to get CraftSpecs.
/// - The class method ClassGetCraftSpecs() is used for instantiation.
///
class Aircraft : public Craft
{
PREVENT_COPYING(Aircraft)
typedef Craft Parent;
friend class shptr<Aircraft>; friend class shptr0<Aircraft>;
public:
	enum eType
	{
		// And update Lua and gui/gui.cc
		eType_Spitfire,
		eType_Corsair,
		eType_P51,
		eType_F82,
		eType_F86,
		eType_A4,
		eType_F15,
		eType_F16,
		eType_F18,
		eType_SR71,
		eType_Mirage2000,
		eType_SU37,
		eType_Sikorsky,
	};

public:
	virtual void	SetState( const eState state );
	virtual void	PlayEngineSound( const fp volume );
	virtual void	StopEngineSound( void );
	virtual void	SetLandingGear( const bool down );

protected:
					// Only derivatives can access Aircraft ctor.
					Aircraft( shptr<Graph> graph, const WorldVertex& pos,
                              const CraftSpecs& craftSpecs );
					~Aircraft();
	virtual void	Tick( const Milliseconds millisecElapsed );
	virtual fp		GetHeight( void );
	bool			IfFlying( void );
	bool			IfCanSetLandingGear( const bool down );

private:
	void			MoreCrashDetermination( void );

public:
	virtual void				Reset( void );
	virtual const string		GetName( void ) { return string("Aircraft"); }
	virtual const CraftSpecs&	GetCraftSpecs( void ) = 0;
	virtual bool				IfHasPropeller( void ) = 0;
	virtual bool				IfHasGuns( void ) { return true; }

public:
	DECL_TYPESIG(TYPESIG_AIRCRAFT)
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Spitfire (Seafire IIIC).
///
class SpitfireAircraft : public Aircraft
{
PREVENT_COPYING(SpitfireAircraft)
typedef Aircraft Parent;
public:
					SpitfireAircraft( shptr<Graph> graph, const WorldVertex& pos );
					~SpitfireAircraft();
	virtual const string GetName( void ) { return string("SpitfireAircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return true; }

private:
	virtual void	Tick( const Milliseconds millisecElapsed );
	void			RotatePropeller( void );

private:
	shptr<Subgraph>		mSubgraphPropeller1, mSubgraphPropeller2;  // 4 blades
	shptr<Subgraph>		mSubgraphPropeller3, mSubgraphPropeller4;
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F4U Corsair.
///
class CorsairAircraft : public Aircraft
{
PREVENT_COPYING(CorsairAircraft)
typedef Aircraft Parent;
public:
					CorsairAircraft( shptr<Graph> graph, const WorldVertex& pos );
					~CorsairAircraft();
	virtual const string GetName( void ) { return string("CorsairAircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return true; }

private:
	virtual void	Tick( const Milliseconds millisecElapsed );
	void			RotatePropeller( void );

private:
	shptr<Subgraph>		mSubgraphPropeller0, mSubgraphPropeller1;
	shptr<Subgraph>		mSubgraphGear0,  mSubgraphGear1,  mSubgraphGear2,  mSubgraphGear3;
	shptr<Subgraph>		mSubgraphGear4,  mSubgraphGear5,  mSubgraphGear6,  mSubgraphGear7;
	shptr<Subgraph>		mSubgraphGear8,  mSubgraphGear9,  mSubgraphGear10, mSubgraphGear11;
	shptr<Subgraph>		mSubgraphGear12, mSubgraphGear13, mSubgraphGear14, mSubgraphGear15;
	shptr<Subgraph>		mSubgraphGear16, mSubgraphGear17, mSubgraphGear18, mSubgraphGear19;
	shptr<Subgraph>		mSubgraphGear20;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief P-51D Mustang.
///
class P51Aircraft : public Aircraft
{
PREVENT_COPYING(P51Aircraft)
typedef Aircraft Parent;
public:
					P51Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~P51Aircraft();
	virtual const string GetName( void ) { return string("P51Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return true; }

private:
	virtual void	Tick( const Milliseconds millisecElapsed );
	void			RotatePropeller( void );

private:
	shptr<Subgraph>		mSubgraphPropeller1;	///< this particular P-51 model
	shptr<Subgraph>		mSubgraphPropeller2;	///< has two nodes for a 4 bladed propeller
	//
	shptr<Subgraph>		mSubgraphLeftGearWheel;
	shptr<Subgraph>		mSubgraphLeftGearStrut;
	shptr<Subgraph>		mSubgraphLeftInnerGearCover;
	shptr<Subgraph>		mSubgraphLeftOuterGearCover;
	shptr<Subgraph>		mSubgraphLeftTailGearDoor;
	//
	shptr<Subgraph>		mSubgraphRightGearWheel;
	shptr<Subgraph>		mSubgraphRightGearStrut;
	shptr<Subgraph>		mSubgraphRightInnerGearCover;
	shptr<Subgraph>		mSubgraphRightOuterGearCover;
	shptr<Subgraph>		mSubgraphRightTailGearDoor;
	//
	shptr<Subgraph>		mSubgraphTailGearWheel;
	shptr<Subgraph>		mSubgraphTailGearStrut;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F-82 Twin Mustang.
///
class F82Aircraft : public Aircraft
{
PREVENT_COPYING(F82Aircraft)
typedef Aircraft Parent;
public:
					F82Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~F82Aircraft();
	virtual const string GetName( void ) { return string("F82Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return true; }

private:
	virtual void	Tick( const Milliseconds millisecElapsed );
	void			RotatePropeller( void );

private:
	shptr<Subgraph>		mSubgraphPropeller1;  // each propeller of the Twin Mustang
	shptr<Subgraph>		mSubgraphPropeller2;  // has 2 subgraphs
	shptr<Subgraph>		mSubgraphPropeller3;
	shptr<Subgraph>		mSubgraphPropeller4;
	//
	shptr<Subgraph>		mSubgraphLeftGearWheel;
	shptr<Subgraph>		mSubgraphLeftGearStrut;
	shptr<Subgraph>		mSubgraphLeftInnerGearCover;
	shptr<Subgraph>		mSubgraphLeftInnerGearCover2;
	shptr<Subgraph>		mSubgraphLeftOuterGearCover;
	shptr<Subgraph>		mSubgraphLeftTailGearDoor;
	//
	shptr<Subgraph>		mSubgraphRightGearWheel;
	shptr<Subgraph>		mSubgraphRightGearStrut;
	shptr<Subgraph>		mSubgraphRightInnerGearCover;
	shptr<Subgraph>		mSubgraphRightInnerGearCover2;
	shptr<Subgraph>		mSubgraphRightOuterGearCover;
	shptr<Subgraph>		mSubgraphRightTailGearDoor;
	//
	shptr<Subgraph>		mSubgraphTailGearWheel;
	shptr<Subgraph>		mSubgraphTailGearStrut;
	shptr<Subgraph>		mSubgraphTailGearWheel2;
	shptr<Subgraph>		mSubgraphTailGearStrut2;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F-86 Sabre.
///
class F86Aircraft : public Aircraft
{
PREVENT_COPYING(F86Aircraft)
typedef Aircraft Parent;
public:
					F86Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~F86Aircraft();
	virtual const string GetName( void ) { return string("F86Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3, mSubgraphGear4, mSubgraphGear5;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief A-4 Skyhawk.
///
class A4Aircraft : public Aircraft
{
PREVENT_COPYING(A4Aircraft)
typedef Aircraft Parent;
public:
					A4Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~A4Aircraft();
	virtual const string GetName( void ) { return string("A4Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3, mSubgraphGear4, mSubgraphGear5;
	shptr<Subgraph>		mSubgraphGear6, mSubgraphGear7, mSubgraphGear8;
	shptr<Subgraph>		mSubgraphGear9, mSubgraphGear10, mSubgraphGear11;
	shptr<Subgraph>		mSubgraphGear12, mSubgraphGear13, mSubgraphGear14;
	shptr<Subgraph>		mSubgraphGear15, mSubgraphGear16, mSubgraphGear17;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F-15 Eagle.
///
class F15Aircraft : public Aircraft
{
PREVENT_COPYING(F15Aircraft)
typedef Aircraft Parent;
public:
					F15Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~F15Aircraft();
	virtual const string GetName( void ) { return string("F15Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3, mSubgraphGear4, mSubgraphGear5;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F-16 Falcon.
///
class F16Aircraft : public Aircraft
{
PREVENT_COPYING(F16Aircraft)
typedef Aircraft Parent;
public:
					F16Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~F16Aircraft();
	virtual const string GetName( void ) { return string("F16Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0,  mSubgraphGear1,  mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3,  mSubgraphGear4,  mSubgraphGear5;
	shptr<Subgraph>		mSubgraphGear6,  mSubgraphGear7,  mSubgraphGear8;
	shptr<Subgraph>		mSubgraphGear9,  mSubgraphGear10, mSubgraphGear11;
	shptr<Subgraph>		mSubgraphGear12, mSubgraphGear13, mSubgraphGear14;
	shptr<Subgraph>		mSubgraphGear15, mSubgraphGear16, mSubgraphGear17;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1, mSubgraphFlame2;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief F-18 Hornet.
///
class F18Aircraft : public Aircraft
{
PREVENT_COPYING(F18Aircraft)
typedef Aircraft Parent;
public:
					F18Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~F18Aircraft();
	virtual const string GetName( void ) { return string("F18Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3, mSubgraphGear4, mSubgraphGear5;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief SR-71 Blackbird.
///
class SR71Aircraft : public Aircraft
{
PREVENT_COPYING(SR71Aircraft)
typedef Aircraft Parent;
public:
					SR71Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~SR71Aircraft();
	virtual const string GetName( void ) { return string("SR71Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual bool	IfHasGuns( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphStrut0,  mSubgraphStrut1,  mSubgraphStrut2,  mSubgraphStrut3;
	shptr<Subgraph>		mSubgraphStrut4,  mSubgraphStrut5,  mSubgraphStrut6,  mSubgraphStrut7;
	shptr<Subgraph>		mSubgraphStrut8,  mSubgraphStrut9,  mSubgraphStrut10, mSubgraphStrut11;
	shptr<Subgraph>		mSubgraphStrut12, mSubgraphStrut13, mSubgraphStrut14, mSubgraphStrut15;
	shptr<Subgraph>		mSubgraphParachute;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1, mSubgraphFlame2;
	shptr<Subgraph>		mSubgraphFlame3, mSubgraphFlame4, mSubgraphFlame5;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Mirage 2000.
///
class Mirage2000Aircraft : public Aircraft
{
PREVENT_COPYING(Mirage2000Aircraft)
typedef Aircraft Parent;
public:
					Mirage2000Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~Mirage2000Aircraft();
	virtual const string GetName( void ) { return string("Mirage2000Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0, mSubgraphGear1, mSubgraphGear2;
	shptr<Subgraph>		mSubgraphGear3, mSubgraphGear4, mSubgraphGear5;
	shptr<Subgraph>		mSubgraphGear6, mSubgraphGear7, mSubgraphGear8;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief SU-37 Flanker.
///
class SU37Aircraft : public Aircraft
{
PREVENT_COPYING(SU37Aircraft)
typedef Aircraft Parent;
public:
					SU37Aircraft( shptr<Graph> graph, const WorldVertex& pos );
					~SU37Aircraft();
	virtual const string GetName( void ) { return string("SU37Aircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfHasPropeller( void ) { return false; }
	virtual void	SetThrottle( const fp throttle );

private:
	shptr<Subgraph>		mSubgraphGear0;
	shptr<Subgraph>		mSubgraphFlame0, mSubgraphFlame1;
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Sikorsky-76 rescue helicopter.
///
class SikorskyAircraft : public Aircraft
{
PREVENT_COPYING(SikorskyAircraft)
typedef Aircraft Parent;
public:
					SikorskyAircraft( shptr<Graph> graph, const WorldVertex& pos );
					~SikorskyAircraft();
	virtual const string GetName( void ) { return string("SikorskyAircraft"); }
	CLASS_METHOD const CraftSpecs& ClassGetCraftSpecs( void );
	virtual      const CraftSpecs&      GetCraftSpecs( void );
	virtual void	SetLandingGear( const bool down );
	virtual bool	IfLandingGear( void ) { return false; }
	virtual bool	IfHasPropeller( void ) { return false; }

private:
	virtual void	Tick( const Milliseconds millisecElapsed );
	void			RotateRotors( void );

private:
	shptr<Subgraph>		mSubgraphRotor0, mSubgraphRotor1, mSubgraphRotor2, mSubgraphRotor3;
	shptr<Subgraph>		mSubgraphRotor4, mSubgraphRotor5, mSubgraphRotor6, mSubgraphRotor7;
	shptr<Subgraph>		mSubgraphTailRotor;
};

shptr<Aircraft>
MakeAircraft( Aircraft::eType aircraftType, const WorldVertex& pos, const bool loadCopy );

} // namespace program

#endif // PROGRAM_AIRCRAFT_HH
