/*****************************************************************************
 * $LastChangedDate: 2008-08-31 02:29:07 -0400 (Sun, 31 Aug 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Updates HUD every tick.
 * @remarks Previously, HudUpdate had slow and fast timer-tick handlers.
 *          The slow handler was reimplemented in Lua.
 *          The fast handler remains in C++ to render the splash text.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define PROGRAM_HUD_CC 1
#include "base/module.hh"
using namespace base;
#include "types/module.hh"
#include "types/rgba_colors.hh"
using namespace types;
#include "hud/module.hh"
using namespace hud;
#include "program/module.hh"
#include "program/hud.hh"

namespace program {

INTERN const fp HUD_COLOR_ALPHA = 1.0f;
INTERN TinyArray<4,RGBA> sSplashGradientColors;
INTERN const Milliseconds SPLASH_FULL_ALPHA_MILLISECONDS( 1500 );
INTERN const Milliseconds SPLASH_FADE_ALPHA_MILLISECONDS( 4500 );

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  HudUpdater  ////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

HudUpdater* HudUpdater::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
HudUpdater::HudUpdater( void )
:	mHud(Hud::GetInstance()),
	mHudStringSplash(CtorHudStringSplash())
{
	mHud.AttachHudString( mHudStringSplash );
}

shptr<Hud::HudString>
HudUpdater::CtorHudStringSplash( void )
{
	Hud::HudString::Args hudStringArgs;
	hudStringArgs.mPos[XX] = 16;
	hudStringArgs.mPos[YY] = 16;
	hudStringArgs.mFontSize = 28.0f;

	// Splash line (center).
	hudStringArgs.mText = defs::PROGRAM_NAME_LONG;
	hudStringArgs.mPos[XX] = 0;
	hudStringArgs.mPos[YY] = 128;  // move upward a little bit
	hudStringArgs.mFontSize = 40.0f;
	hudStringArgs.mOrientation = Hud::HudString::eOrientation_CENTER;
	shptr<Hud::HudString> hudStringSplash = new Hud::HudString( hudStringArgs );
	sSplashGradientColors[Hud::HudString::eGradientColor_TOP_LEFT]     = \
	sSplashGradientColors[Hud::HudString::eGradientColor_TOP_RIGHT]    = RGBA( 0.0f, 0.8f, 1.0f, 1.0f );
	sSplashGradientColors[Hud::HudString::eGradientColor_BOTTOM_LEFT]  = \
	sSplashGradientColors[Hud::HudString::eGradientColor_BOTTOM_RIGHT] = RGBA( 0.0f, 0.2f, 1.0f, 1.0f );
	hudStringSplash->SetGradientColors4( sSplashGradientColors );
	return hudStringSplash;
}

HudUpdater::~HudUpdater()
{
	// NOP
}

/*****************************************************************************
 * Start the updating of HUD.
 *****************************************************************************/
void
HudUpdater::Start( void )
{
BEGIN_ONCE {
	Timer::GetInstance().Register( new TickFunctor,
                                   defs::HUD_UPDATER_FAST_TICK_FREQ,
                                   Timer::CANNOT_DISABLE );
} END_ONCE
}

/*****************************************************************************
 * Pulse.
 *****************************************************************************/
void
HudUpdater::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	// Reflect.
	HudUpdater::GetInstance().UpdateFast( millisecElapsed );
}

/*****************************************************************************
 * This is the fast-tick handler (splash).
 *****************************************************************************/
void
HudUpdater::UpdateFast( const Milliseconds millisecElapsed )
{
	// Splash line.
	// For a few seconds, keep alpha at 1.0.
	// Then for the next few seconds, fade-out the splash.
	uint8 splashAlpha = 0;
	if ( millisecElapsed < SPLASH_FULL_ALPHA_MILLISECONDS )
	{
		splashAlpha = 0xff;
	}
	else if ( millisecElapsed < SPLASH_FADE_ALPHA_MILLISECONDS )
	{
		splashAlpha = uint8( 0xff - 255.0f * fp(millisecElapsed) / fp(SPLASH_FADE_ALPHA_MILLISECONDS) );
	}
	sSplashGradientColors[Hud::HudString::eGradientColor_TOP_LEFT][AA]     = \
	sSplashGradientColors[Hud::HudString::eGradientColor_TOP_RIGHT][AA]    = \
	sSplashGradientColors[Hud::HudString::eGradientColor_BOTTOM_LEFT][AA]  = \
	sSplashGradientColors[Hud::HudString::eGradientColor_BOTTOM_RIGHT][AA] = splashAlpha;
	mHudStringSplash->SetGradientColors4( sSplashGradientColors );
}

} // namespace program
