/*****************************************************************************
 * $LastChangedDate: 2008-09-13 12:35:14 -0400 (Sat, 13 Sep 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	World class.
 *//*
 * LEGAL:   COPYRIGHT (C) 2007 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define WORLD_WORLD_CC 1
#define NEED_PLANET_CLASS 1
#include "base/module.hh"
#include "base/conf.hh"
using namespace base;
#include "glue/module.hh"
#include "glue/osg.hh"
using namespace glue;
#include "object/module.hh"
using namespace object;
#include "graph/module.hh"
using namespace graph;
#include "collision/module.hh"
using namespace collision;
#include "view/module.hh"
using namespace view;
#include "sky/module.hh"
#include "sky/conf.hh"
using namespace sky;
#include "world/module.hh"
#include "world/defs.hh"
#include "world/conf.hh"
#include "world/planet.hh"

namespace world {

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////  World  ///////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * Make() is the actual ctor.
 *****************************************************************************/
World::World( void )
:	mRootGroup(new GroupNode),
	mStateSorter(new StateSorter),
	mZombies()
{
	// NOP
}

World::~World()
{
	// NOP
}

/*****************************************************************************
 * Called by world::Init() [world/module.cc].
 *****************************************************************************/
void
World::Make( void )
{
BEGIN_ONCE {

	// Attach to scene.
	GET_SCENE_GRAPH()->AttachNode( mRootGroup.get() );

	// Set the clear/background color.
	if ( WORLD_CONF.mFlat )
	    GET_SCENE_GRAPH()->SetClearColor( RGBA(0.1f,0.1f,0.1f,1.0f) );
	else  // [2008/04] sky not fully implemented for OSSIM
		GET_SCENE_GRAPH()->SetClearColor( SKY_CONF.mSkyColor );

	// Create and attach Planet.  OSSIM is too dark with lighting.
	shptr<Object> planet = GET_PLANET().GetRootObject();
	planet->SetCollidable( true );  // planet is a collidable object
	Attach( planet, WORLD_CONF.mFlat ? STATE_BIT_DEFAULT : STATE_BIT_NO_LIGHTING );

	// Register timer-tick handler to destroy zombies.
	Timer::GetInstance().Register( new TickFunctor, ZOMBIE_TICK_FREQ );

} END_ONCE
}

/*****************************************************************************
 * Attach Object to World under appropriate sub-graph.
 *
 * The pair of stateBits,stateSet is somewhat of an idiom.
 * stateBits is a value that exists on the app-side to state-sort
 * the nodes of the scene-graph.
 * stateSet is a object that affects graphical state (applied by OSG).
 * Thru StateSorter, stateBits affects osg::StateSet.
 * The caller can pass a osg::StateSet arg modified in ways not possible by StateBits.
 *
 * @param	object
 * @param	stateBits
 *			Value that represents a graphical state (osg::StateSet).
 * @param	stateSet
 *			Actual graphical state (osg::StateSet).
 *          This is optional (if not passed, then new osg::StateSet is compiled).
 *****************************************************************************/
void
World::Attach( shptr<Object> object,
               const StateBits stateBits,
               RefPtr<osg::StateSet> stateSet )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

    // Get Object's node.
    RefPtr<Node> objectNode = object->GetGraph()->GetRootNode();

    // Ensure it is unattached.
    if ( UX( objectNode->getNumParents() > 0 ) )
        Detach( object );

	// For optimal state-sorting.
	RefPtr<GroupNode> sortedGroup = \
		mStateSorter->GetSortedGroup( mRootGroup, stateBits, stateSet );
	sortedGroup->addChild( objectNode.get() );

	// Optimize this Object's graph.
	GET_SCENE_GRAPH()->OptimizeNode( objectNode );
}

/*****************************************************************************
 * Detach Object from World.
 * @post Detached Object will be marked non-collidable and probably destroyed.
 *****************************************************************************/
void
World::Detach( shptr<Object> object )
{
CHECK_TYPESIG(object,TYPESIG_OBJECT);

    // An Object should be attached to 0 or 1 group nodes.
   	RefPtr<osg::Node> rootNode = object->GetRootNode();
    const uint parentCnt = rootNode->getNumParents();
    if ( parentCnt == 1 )
    {
    	// Has one parent which should be an osg::Group.
        RefPtr<osg::Group> parent = rootNode->getParent( 0 );  // idx=0
    #if DEBUG
        ASSERT( parent->asGroup() != NULL );                // should be a Group or derivative
        ASSERT( parent->removeChild( rootNode.get() ) );    // remove() should return true
    #else
        parent->removeChild( rootNode.get() );
    #endif
    }
    else // parentCnt != 1
    {
        // Multiple parents shouldn't occur and isn't handled.
        // Objects may have shared nodes for 3D models but
        // those should be underneath unique group nodes (osg::MatrixTransform).
        ASSERT( parentCnt == 0 );
    }

	// And release Collision's reference to this Object.
	// Rationale is that Detach() is called to free an Object.
	object->SetCollidable( false );

    ASSERT( rootNode->getNumParents() == 0 );  // should have zero parents
}

/*****************************************************************************
 * Optimize entire scene-graph.
 *****************************************************************************/
void
World::OptimizeSceneGraph( void )
{
	GET_SCENE_GRAPH()->OptimizeSceneGraph();
}

/*****************************************************************************
 * If the World is flat ;-)
 * This is queried since FlatPlanet is deficient and some configuration
 * is only applicable to FlatPlanet.
 *****************************************************************************/
bool
World::IfFlat( void )
{
	return GET_PLANET().IfFlat();
}

/*****************************************************************************
 * Get radius of World/Planet.
 *****************************************************************************/
fp
World::GetRadius( void )
{
	return GET_PLANET().RadiusGraphical();
}

/*****************************************************************************
 * Set compression of World.
 * @returns Previous flag.
 *****************************************************************************/
bool
World::SetCompression( const bool enable )
{
	return GET_PLANET().SetCompression( enable );
}

/*****************************************************************************
 * Get compression scale of World.
 *****************************************************************************/
fpx
World::GetCompressionScale( void )
{
	return GET_PLANET().GetCompressionScale();
}

/*****************************************************************************
 * Orient an Object to be tangent to a point on the world (in case world is spherical).
 *****************************************************************************/
void
World::RotateTangential( Object& object, const SphereVertex& sv )
{
	if ( not GET_PLANET().IfFlat() )
	{
		Matrix m = GetTangentMatrix( sv );
		m.setTrans( object.GetPosition() );  // keep Object's position
		MatrixRotateLocal( m, XX, Deg2Rad(90.0f) );  // otherwise Object points up (normal to surface)
		object.SetMatrix( m );
	}
	// else flat so NOP
}

/*****************************************************************************
 * Compute a matrix that is tangent to a point on the world.
 *****************************************************************************/
const Matrix&
World::GetTangentMatrix( const SphereVertex& sv )
{
	return GET_PLANET().GetTangentMatrix( sv );
}

/*****************************************************************************
 * Compute a normalized "up vector" extending from a point on the world.
 * A "down vector" can be computed by multipling by -1.0f.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( const SphereVertex& sv )
{
	return GET_PLANET().ComputeUpVector( sv );
}

/*****************************************************************************
 * Compute an "up vector" where an Object is at.
 *****************************************************************************/
WorldVertex
World::ComputeUpVector( Object& object )
{
CHECK_TYPESIG(&object,TYPESIG_OBJECT);

	return ComputeUpVector( Geo::WorldVertex2SphereVertex( object.GetPosition() ) );
}

/*****************************************************************************
 * @return Volume of world in 3D space.
 *****************************************************************************/
BoxVolume
World::GetVolume( void )
{
	return GET_PLANET().GetVolume();
}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////  World : zombies  ////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Schedule deletion of an Object.
 * Zombies are used to allow an object to destroy itself, and also
 * are used to avoid disrupting an iteration by delaying deletions.
 *****************************************************************************/
void
World::SpawnZombie( shptr<Object> object )
{
ASSERT_WORLD();

	// Duplicate zombies are possible but shptr prevents double-deletion.
	mZombies.push_back( object );
}

/*****************************************************************************
 * Delete zombie Objects by detaching them out of STL container.
 * shptr will delete Objects that are no longer referenced.
 *****************************************************************************/
void
World::KillZombies( void )
{
ASSERT_WORLD();

	// Iterate thru STL vector, detach every Zombie out of World.
	// Detach() removes a shptr which will auto-delete unreferenced Objects.
	for ( Zombies::iterator iter = mZombies.begin(); iter != mZombies.end(); ++iter )
	{
		shptr<Object> object = *iter;  CHECK_TYPESIG(object,TYPESIG_OBJECT);
		Detach( object );
	}

	mZombies.clear();
}

// Reflect.
void
World::TickFunctor::operator()( const Milliseconds millisecElapsed )
{
	gWorld.KillZombies();
}

} // namespace world
